"""
Created on Sep 17, 2014

"""
import codecs
import csv
import logging
import os
from typing import *
from typing import TextIO
from pathlib import Path

from itertools import islice

from bi_etl.components.etlcomponent import ETLComponent
from bi_etl.scheduler.task import ETLTask
from bi_etl.timer import Timer

__all__ = ['CSVReader']

# Only quote none is really needed here, QUOTE_MINIMAL is the default.
# The other quoting levels are only relevant to the Writer
QUOTE_NONE = csv.QUOTE_NONE
QUOTE_MINIMAL = csv.QUOTE_MINIMAL


class CSVReader(ETLComponent):
    """ 
    CSVReader is similar to csv.DictReader
    However, instead of a dict it uses our :class:`~bi_etl.components.row.row.Row` class as it's return type.
    It uses :class:`csv.reader` (in :mod:`csv`) to read the file.
    
    Note optional, but important, parameter ``delimiter``.
    
    **Valid values for `errors` parameter:**
             
    +--------------------+-----------------------------------------------------------+
    | Value              | Meaning                                                   |
    +====================+===========================================================+
    | 'strict'           | raise a ValueError error (or a subclass)                  |
    +--------------------+-----------------------------------------------------------+
    | 'ignore'           | ignore the character and continue with the next           |
    +--------------------+-----------------------------------------------------------+
    | 'replace'          | replace with a suitable replacement character;            |
    |                    | Python will use the official U+FFFD REPLACEMENT           |
    |                    | CHARACTER for the builtin Unicode codecs on               |
    |                    | decoding and '?' on encoding.                             |
    +--------------------+-----------------------------------------------------------+
    | 'surrogateescape'  | replace with private code points U+DCnn.                  |
    +--------------------+-----------------------------------------------------------+               
    
    Args:
        task: The  instance to register in (if not None)
        
        filedata: 
            The file to parse as delimited. If str then it's assumed to be a filename. 
            Otherwise it's assumed to be a file object.
    
        encoding: 
            The encoding to use when opening the file, 
            if it was a filename and not already opened.
            Default is None which becomes the Python default encoding 
            
        errors: 
            The error handling to use when opening the file 
            (if it was a filename and not already opened)
            Default is 'strict'
            See above for valid errors values.            
            
        logical_name:
            The logical name of this source. Used for log messages.  
    
    Attributes:
        column_names: list
            The names to use for columns
            
        primary_key: str
            The name of the primary key column. Only impacts trace messages.  Default=None.
        
        dialect: str or subclass of :class:`csv.Dialect`
            Default "excel". The dialect value to pass to :mod:`csv`
        
        delimiter: str
            The delimiter used in the file. Default is ','.
        
        doublequote: boolean
            Controls how instances of quotechar appearing inside a column should themselves be quoted. 
            When True, the character is doubled. 
            When False, the escapechar is used as a prefix to the quotechar. 
            It defaults to True.
        
        escapechar: str
            A one-character string used by the writer to escape the delimiter if quoting is set to 
            QUOTE_NONE and the quotechar if doublequote is False. On reading, the escapechar removes
            any special meaning from the following character. It defaults to None, which disables escaping.
        
        quotechar: str
            A one-character string used to quote columns containing special characters, such as the delimiter
            or quotechar, or which contain new-line characters. 
            It defaults to '"'.
        
        quoting: 
            Controls when quotes should be generated by the writer and recognised by the reader.
            Can be either of the constants defined in this module.
            * QUOTE_NONE 
            * QUOTE_MINIMAL
            Defaults to QUOTE_MINIMAL.
            
            For more details see https://docs.python.org/3/library/csv.html#csv.QUOTE_ALL
        
        skipinitialspace: boolean
            When True, whitespace immediately following the delimiter is ignored. The default is False.
        
        strict: boolean
            When True, raise exception Error on bad CSV input. The default is False.
    
        header_row: int
            The row to parse for headers
        
        start_row: int
            The first row to parse for data
            
        log_first_row : boolean
            Should we log progress on the the first row read. *Only applies if Table is used as a source.*
            (inherited from ETLComponent)
            
        max_rows : int, optional
            The maximum number of rows to read. *Only applies if Table is used as a source.*
            (inherited from ETLComponent)
            
        primary_key: list
            The name of the primary key column(s). Only impacts trace messages.  Default=None.
            (inherited from ETLComponent)
        
        progress_frequency: int
            How often (in seconds) to output progress messages. None for no progress messages.
            (inherited from ETLComponent)
        
        progress_message: str
            The progress message to print. Default is ``"{logical_name} row # {row_number}"``. 
            Note ``logical_name`` and ``row_number`` subs.
            (inherited from ETLComponent)
            
        restkey: str
            Column name to catch long rows (extra values).
            
        restval: str
            The value to put in columns that are in the column_names but 
            not present in a given row (missing values).  
            
        large_field_support: boolean
            Enable support for csv columns bigger than 131,072 default limit.
    """

    DELIMITER_SNIFF_LINES = 10

    def __init__(self,
                 task: Optional[ETLTask],
                 filedata: Union[TextIO, str, Path],
                 encoding: str = None,
                 errors: str = 'strict',
                 logical_name: str = None,
                 **kwargs
                 ):
        
        self.log = logging.getLogger(f"{self.__class__.__module__}.{self.__class__.__name__}")
        
        self.__close_file = False
        
        if isinstance(filedata, Path):
            filedata = str(filedata)
        
        # We have to check / open the file here to get the name for the logical name
        if isinstance(filedata, str):
            self.log.debug(f"Opening file {filedata}")

            if encoding is None or encoding in {'utf8', 'utf-8'}:
                with open(filedata, 'rb') as binary_file:
                    begin_of_file = binary_file.read(3)
                    if begin_of_file == codecs.BOM_UTF8:
                        encoding = 'utf-8-sig'
                        self.log.warning(f'BOM found for {filedata} using encoding {encoding}')
                    else:
                        self.log.debug(f'No BOM found for {filedata} ')

            self.file = open(filedata, 
                             mode='rt',
                             newline='', 
                             encoding=encoding, 
                             errors=errors
                             )
            self.__close_file = True            
        else:
            self.log.info("Treating input as file object {}".format(filedata))
            self.file = filedata
        
        if logical_name is None:
            try: 
                logical_name = os.path.basename(self.file.name)
            except AttributeError:
                logical_name = str(self.file)
        
        # Don't pass kwargs up. They should be set here at the end
        super(CSVReader, self).__init__(task=task,
                                        logical_name=logical_name,
                                        )
        self.__reader = None
        
        # Begin csv module params        
        self.dialect = "excel"
        self.delimiter = None
        self.doublequote = True
        self.escapechar = None
        self.quotechar = '"'
        self.quoting = csv.QUOTE_MINIMAL
        self.skipinitialspace = False
        self.strict = False
        self.large_field_support = False
        # End csv module params
        self.header_row = 1
        self.start_row = 1  # Will be incremented if the header row is read
        
        # column to catch long rows (more values than columns)
        self.restkey = 'extra data past last delimiter'
        # default value for short rows (value for missing keys)    
        self.restval = None   
        self.trace_data = False
        self.extra_part_msg_limit = 100
        self.extra_part_msg_cnt = 0
        
        # Should be the last call of every init
        self.set_kwattrs(**kwargs)

    def __repr__(self):
        return "{cls}(task={task},logical_name={logical_name},filedata={file},primary_key={primary_key}," \
               "column_names={column_names})".format(
                    cls=self.__class__.__name__,
                    task=self.task,
                    logical_name=self.logical_name,
                    file=self.file,
                    primary_key=self.primary_key,
                    column_names=self.column_names,
                    )

    @property
    def reader(self) -> csv.reader:
        """
        Build or get the csv.reader object.
        """
        if self.__reader is None:
            if self.delimiter is None:
                delimiters = ''.join([',', '|', '\t'])
                try:
                    # Note: The sniff routine builds a histogram for delimiter frequency
                    #       So we want to make sure to give it whole lines and not cut the
                    #       last line off in the middle.
                    sample_data = ''.join(islice(self.file, self.DELIMITER_SNIFF_LINES))
                    dialect = csv.Sniffer().sniff(sample_data, delimiters=delimiters)
                except csv.Error as e:
                    try:
                        name = self.file.name
                    except AttributeError:
                        name = self.file
                    raise ValueError(f"{e} in file {name} delimiters={delimiters}")
                if dialect.delimiter not in delimiters:
                    msg = f'Invalid delimiter \'{dialect.delimiter}\' found in {self.file.name} csv file. ' \
                          f'Expected one of {delimiters}'
                    self.log.warning(msg)
                    raise ValueError(msg)
                else:
                    self.log.info(f'Found delimiter \'{dialect.delimiter}\' in {self.file.name} csv file.')
                self.dialect = dialect

                self.file.seek(0)

                self.delimiter = dialect.delimiter

            # Initialize the reader
            self.__reader = csv.reader(
                self.file,
                dialect=self.dialect,
                delimiter=self.delimiter,
                doublequote=self.doublequote,
                escapechar=self.escapechar,
                quotechar=self.quotechar,
                quoting=self.quoting,
                skipinitialspace=self.skipinitialspace,
                strict=self.strict,
                )
            if self.large_field_support:
                csv.field_size_limit(2147483647)  # largest value it will accept
            
        return self.__reader

    @property
    def line_num(self):
        """
        The current line number in the source file.
        line_num differs from rows_read in that rows_read deals with rows that would be returned to the caller
        """
        return self.reader.line_num
    
    def seek_row(self, target_row):
        saved_position = None
        saved_line_num = None        
        if target_row is not None:
            current_row = self.line_num + 1
            if target_row < current_row:
                if self.file.seekable():
                    saved_position = self.file.tell()
                    saved_line_num = self.reader.line_num
                    self.file.seek(0)
                    try:
                        self.reader.line_num = 0
                    except AttributeError:
                        pass                    
                else:
                    raise ValueError("Un-seekable file, already read past the seek target row {:,}".format(target_row))
            elif target_row > current_row:                
                self.log.info("Seeking row {:,}".format(target_row))
                stats = self.get_stats_entry('seek')
                stats.timer.start()
                progress_timer = Timer()                                
                for _ in range(target_row - current_row):
                    # noinspection PyTypeChecker
                    next(self.reader, None)
                    # Increment self.rows_read although they aren't really processed, so we know where we were
                    stats['rows read'] += 1
                    stats['lines read'] = self.line_num
                    if self.progress_frequency is not None: 
                        if progress_timer.seconds_elapsed >= self.progress_frequency:
                            self.log.info("Seek reached row {:,}".format(stats['rows read']))
                            progress_timer.reset()
                self.log.info("Done seeking row {:,}".format(self.start_row))
        return saved_position, saved_line_num
        
    def __read_header_row(self):
        """
        Read the header row. This is a function so it can be overridden with more
        complex header parsing.
        """
        (saved_position, saved_line_num) = self.seek_row(self.header_row)
        if self.start_row == self.header_row:
            self.start_row += 1
        # noinspection PyTypeChecker
        header_row = next(self.reader, None)
        if saved_position is not None and saved_position > 0:
            new_position = self.file.tell()
            if new_position < saved_position:
                self.file.seek(saved_position)
            self.reader.line_num = saved_line_num
        return header_row
    
    def _obtain_column_names(self):
        """
        Get the column names from the file. ETLComponent only call this if self._column_names is None:
        """
        try:            
            header_row = self.__read_header_row()
            assert header_row is not None, "Header row not found (or empty)"
            # Make sure to use setter here! It deals with duplicate names                
            self.column_names = header_row

        except StopIteration:
            self.column_names = []

        if self.trace_data:
            self.log.debug(f"Column names read: {self.column_names}")
    
    def _raw_rows(self):
        len_column_names = len(self.column_names)
        try:
            self.seek_row(self.start_row)
            this_iteration_header = self.full_iteration_header
            # noinspection PyTypeChecker
            for row in self.reader:
                if len(row) != 0:     
                    # Convert empty strings to None to be consistent with DB reads
                    row = [None if s == '' else s for s in row]
                    d = self.Row(iteration_header=this_iteration_header)
                    d.update_from_values(row[:len_column_names])
                    
                    len_row = len(row)
                    if len_column_names < len_row:
                        # Note: Adding restkey to the row will create a new iteration header
                        #      (shared by subsequent rows with extra values)
                        d[self.restkey] = row[len_column_names:]
                        if self.extra_part_msg_cnt < self.extra_part_msg_limit:
                            self.extra_part_msg_cnt += 1
                            self.log.debug("Extra part of row {:,} read={}".format(self.rows_read, row[len_column_names:]))
                            if self.extra_part_msg_cnt == self.extra_part_msg_limit:
                                self.log.debug("No more Extra part of row read messages will be logged")
                    elif len_column_names > len_row:
                        # This could be done in a faster way, but hopefully is rare so not worth optimizing
                        for key in self.column_names[len_row:]:
                            d[key] = self.restval
                    yield d
        except csv.Error as e:
            if 'field larger than field limit' in str(e):
                raise ValueError('{}. Please enable CSVReader.large_field_support.')
            else:
                raise e
        
    def close(self, error: bool = False):
        """
        Close the file
        """
        if not self.is_closed:
            if self.__close_file:
                self.file.close()
            super(CSVReader, self).close(error=error)



import textwrap
import typing

from bi_etl.components.hst_table import HistoryTable
from bi_etl.components.readonlytable import ReadOnlyTable
from bi_etl.components.row.row import Row
from bi_etl.components.row.row_status import RowStatus
from bi_etl.components.table import Table
from bi_etl.database import DatabaseMetadata
from bi_etl.exceptions import NoResultFound
from bi_etl.scheduler.task import ETLTask
from bi_etl.statistics import Statistics
from bi_etl.timer import Timer

__all__ = ['HistoryTableSourceBased']


# noinspection PyAbstractClass
class HistoryTableSourceBased(HistoryTable):
    """
    ETL target component for a table that stores history of updates where the source provides the dates to use.
    Also usable as a source.

    Example:
        Target table currently has:
            2016 - A
            2017 - B
            2018 - C

        Source Deletes 2017 version.  A normal HistoryTable load would not touch the 2017 version since it was not
        provided in the source.

        The desired output is:
            2016 - A
            2017 - A
            2018 - C

        This component handles that update by stepping through the provided source versions and existing target versions
        updating each target version as required so it matches the closest source version earlier or equal in time.

    Parameters
    ----------
    task : ETLTask
        The  instance to register in (if not None)

    database : bi_etl.scheduler.task.Database
        The database to find the table/view in.

    table_name : str
        The name of the table/view.

    exclude_columns : frozenset
        Optional. A list of columns to exclude from the table/view. These columns will not be included in SELECT, INSERT, or UPDATE statements.

    Attributes
    ----------

    auto_generate_key: boolean
        Should the primary key be automatically generated by the insert/upsert process?
        If True, the process will get the current maximum value and then increment it with each insert.

    auto_commit: boolean
        Automatically commit after delete? Defaults to False.

    begin_date: str
        Name of the begin date field

    end_date: str
        Name of the end date field

    inserts_use_default_begin_date: boolean
        Should inserts use the default begin date instead of the class effective date
        This allows records to match up in joins with other history tables where the effective
        date of the 'primary' might be before the first version effective date.
        Default = True

    default_begin_date: date
        Default begin date to assign for begin_date.
        Used for new records if inserts_use_default_begin_date is True.
        Also used for
        :meth:`get_missing_row`,
        :meth:`get_invalid_row`,
        :meth:`get_not_applicable_row`,
        :meth:`get_various_row`
        Default = 1900-1-1

    default_end_date: date
        Default begin date to assign for end_date for active rows.
        Also used for
        :meth:`get_missing_row`,
        :meth:`get_invalid_row`,
        :meth:`get_not_applicable_row`,
        :meth:`get_various_row`
        Default = 9999-1-1

    auto_generate_key: boolean
        Should the primary key be automatically generated by the insert/upsert process?
        If True, the process will get the current maximum value and then increment it with each insert.
        (inherited from Table)

    batch_size: int
        How many rows should be insert/update/deleted in a single batch.
        (inherited from Table)

    delete_flag : str
        The name of the delete_flag column, if any.
        Optional.
        (inherited from ReadOnlyTable)

    delete_flag_yes : str
        The value of delete_flag for deleted rows.
        Optional.
        (inherited from ReadOnlyTable)

    delete_flag_no : str
        The value of delete_flag for *not* deleted rows.
        Optional.
        (inherited from ReadOnlyTable)

    default_date_format: str
        The date parsing format to use for str -> date conversions.
        If more than one date format exists in the source, then explicit conversions will be required.

        Default = '%m/%d/%Y'
        (inherited from Table)

    force_ascii: boolean
        Should text values be forced into the ascii character set before passing to the database?
        Default = False
        (inherited from Table)

    last_update_date: str
        Name of the column which we should update when table updates are made.
        Default = None
        (inherited from Table)

    log_first_row : boolean
        Should we log progress on the the first row read. *Only applies if Table is used as a source.*
        (inherited from ETLComponent)

    max_rows : int
        The maximum number of rows to read. *Only applies if Table is used as a source.*
        Optional.
        (inherited from ETLComponent)

    primary_key: list
        The name of the primary key column(s). Only impacts trace messages.  Default=None.
        If not passed in, will use the database value, if any.
        (inherited from ETLComponent)

    natural_key: list
        The list of natural key columns (as Column objects).
        The default is the list of non-begin/end date primary key columns.
        The default is *not* appropriate for dimension tables with surrogate keys.

    progress_frequency : int
        How often (in seconds) to output progress messages.
        Optional.
        (inherited from ETLComponent)

    progress_message : str
        The progress message to print.
        Optional. Default is ``"{logical_name} row # {row_number}"``. Note ``logical_name`` and ``row_number``
        substitutions applied via :func:`format`.
        (inherited from ETLComponent)

    special_values_descriptive_columns: list
         A list of columns that get longer descriptive text in
         :meth:`get_missing_row`,
         :meth:`get_invalid_row`,
         :meth:`get_not_applicable_row`,
         :meth:`get_various_row`
         Optional.
         (inherited from ReadOnlyTable)

    track_source_rows: boolean
        Should the :meth:`upsert` method keep a set container of source row keys that it has processed?
        That set would then be used by :meth:`update_not_processed`, :meth:`logically_delete_not_processed`,
        and :meth:`delete_not_processed`.
        (inherited from Table)

    type_1_surrogate: str
        The name of the type 1 surrogate key. The value is automatically generated as equal to the type 2 key on
        inserts and equal to the existing value on updates.
        Optional.

    """

    def __init__(self,
                 task: typing.Optional[ETLTask],
                 database: DatabaseMetadata,
                 table_name: str,
                 table_name_case_sensitive: bool = None,
                 schema: str = None,
                 exclude_columns: frozenset = None,
                 **kwargs
                 ):
        # Don't pass kwargs up to super. They should be set at the end of this method
        super().__init__(task=task,
                         database=database,
                         table_name=table_name,
                         schema=schema,
                         table_name_case_sensitive=table_name_case_sensitive,
                         exclude_columns=exclude_columns,
                         )
        self.prior_upsert_key = None
        self.prior_upsert_lookup_name = None
        self.prior_upsert_existing_rows_list = list()
        self.prior_upsert_existing_row_index = 0
        self.prior_upsert_row = None
        self.prior_upsert_update_callback = None
        self.prior_upsert_insert_callback = None
        self.prior_upsert_stat_name = None
        self.prior_upsert_parent_stats = None

        # Should be the last call of every init
        self.set_kwattrs(**kwargs)

    def _upsert_version_set(
            self,
            source_mapped_as_target_row: Row,
            lookup_name: str = None,
            skip_update_check_on: list = None,
            do_not_update: list = None,
            additional_update_values: dict = None,
            additional_insert_values: dict = None,
            update_callback: typing.Callable[[list, Row], None] = None,
            insert_callback: typing.Callable[[Row], None] = None,
            source_excludes: frozenset = None,
            target_excludes: frozenset = None,
            stat_name: str = 'upsert',
            parent_stats: Statistics = None,
            ):
        """
        Update (if changed) or Insert a row in the table.
        Returns the row found/inserted, with the auto-generated key (if that feature is enabled)

        Parameters
        ----------
        source_mapped_as_target_row: :class:`~bi_etl.components.row.row_case_insensitive.Row`
            Row to upsert
        lookup_name: str
            The name of the lookup (see :meth:`define_lookup`) to use when searching for an existing row.
        skip_update_check_on: list
            List of column names to not compare old vs new for updates.
        do_not_update: list
             List of columns to never update.
        additional_update_values: dict
            Additional updates to apply when updating
        additional_insert_values: dict
            Additional values to set on each row when inserting.
        update_callback: func
            Function to pass updated rows to. Function should not modify row.
        insert_callback: func
            Function to pass inserted rows to. Function should not modify row.
        source_excludes:
            list of source columns to exclude when mapping to this Table.
        target_excludes:
            list of Table columns to exclude when mapping from the source row
        stat_name: string
            Name of this step for the ETLTask statistics. Default = 'upsert'
        parent_stats: bi_etl.statistics.Statistics
            Optional Statistics object to nest this steps statistics in.
            Default is to place statistics in the ETLTask level statistics.
        """
        # This should be the (or a) natural key lookup
        lookup = self.get_lookup(lookup_name)
        nk_tuple = lookup.get_hashable_combined_key(source_mapped_as_target_row)

        if self.prior_upsert_key != nk_tuple or self.prior_upsert_lookup_name != lookup_name:
            # Check if this is the first row and not the end of a batch
            if self.prior_upsert_key is not None:
                # This is the end of a batch
                self.finish_pending_existing_rows()

            # Reset for the next list
            self.prior_upsert_row = None
            self.prior_upsert_key = nk_tuple
            self.prior_upsert_lookup_name = lookup_name
            fallback_to_db = self.always_fallback_to_db or not self.cache_clean
            try:
                existing_rows = lookup.find_versions_list(self.prior_upsert_key, fallback_to_db=fallback_to_db)
            except NoResultFound:
                # No existing rows were found, insert
                new_row = source_mapped_as_target_row
                if self.inserts_use_default_begin_date:
                    new_row.set_keeping_parent(self.begin_date_column, self.default_begin_date)

                if additional_insert_values:
                    for colName, value in additional_insert_values.items():
                        new_row[colName] = value
                new_row = self.insert_row(
                    new_row,
                    parent_stats=parent_stats,
                )

                if insert_callback:
                    insert_callback(new_row)

                return new_row

            # We only reach here if a result was found from the lookup
            self.prior_upsert_existing_rows_list = existing_rows
            self.prior_upsert_existing_row_index = 0
            self.prior_upsert_update_callback = update_callback
            self.prior_upsert_insert_callback = insert_callback
            self.prior_upsert_stat_name = stat_name
            self.prior_upsert_parent_stats = parent_stats
            if self.inserts_use_default_begin_date:
                # Force first row to use the default begin date (year 1900)
                row_begin_date = self.default_begin_date
                source_mapped_as_target_row.set_keeping_parent(self.begin_date_column, row_begin_date)
            else:
                row_begin_date = source_mapped_as_target_row[self.begin_date_column]
        else:
            row_begin_date = source_mapped_as_target_row[self.begin_date_column]

        # Check that rows are coming in order
        if self.prior_upsert_row is not None:
            if row_begin_date < self.prior_upsert_row[self.begin_date_column]:
                raise RuntimeError(
                    textwrap.dedent(f"""\
                    rows not in begin date order!
                    nk_tuple = {nk_tuple}
                    row = {repr(source_mapped_as_target_row)}
                    prior_upsert_row = {repr(self.prior_upsert_row)}
                    """)
                )

        # Nested join over existing rows and pending upsert rows
        first_new_row = None
        get_next_new_row = False
        upsert_row = source_mapped_as_target_row
        while not get_next_new_row:
            needs_upsert = True
            if self.prior_upsert_existing_row_index >= len(self.prior_upsert_existing_rows_list):
                # There is no next existing row, use a max date value
                existing_begin_date = self.default_end_date
                existing_row = None
            else:
                existing_row = self.prior_upsert_existing_rows_list[self.prior_upsert_existing_row_index]
                existing_begin_date = existing_row[self.begin_date_column]

            if row_begin_date < existing_begin_date:
                # New date is before all remaining existing rows
                upsert_row = source_mapped_as_target_row
                get_next_new_row = True
            elif row_begin_date == existing_begin_date:
                # New date equal to existing date
                upsert_row = source_mapped_as_target_row
                get_next_new_row = True
                self.prior_upsert_existing_row_index += 1
            else:
                # New row begin > Existing date
                if self.prior_upsert_row is not None:
                    excludes = set(self.primary_key)
                    excludes.add(self.end_date_column)
                    if source_excludes is not None:
                        excludes.update(source_excludes)
                    upsert_row = self.prior_upsert_row.subset(exclude=excludes)
                    upsert_row.set_keeping_parent(self.begin_date_column, existing_begin_date)
                    self.prior_upsert_existing_row_index += 1
                else:
                    # No prior row, keep existing data
                    upsert_row = existing_row
                    self.prior_upsert_existing_row_index += 1
                    # As the logic stands now we could skip the upsert entirely in this scenario
                    needs_upsert = False

            if needs_upsert:
                new_row = super().upsert(
                    source_row=upsert_row,
                    lookup_name=lookup_name,
                    skip_update_check_on=skip_update_check_on,
                    do_not_update=do_not_update,
                    additional_update_values=additional_update_values,
                    additional_insert_values=additional_insert_values,
                    update_callback=update_callback,
                    insert_callback=insert_callback,
                    source_excludes=source_excludes,
                    target_excludes=target_excludes,
                    stat_name=stat_name,
                    parent_stats=parent_stats,
                    )
                if first_new_row is None:
                    first_new_row = new_row
        self.prior_upsert_row = upsert_row

        return first_new_row

    def finish_pending_existing_rows(self):
        if self.prior_upsert_row is not None:
            for existing_row in self.prior_upsert_existing_rows_list[self.prior_upsert_existing_row_index:]:
                existing_begin_date = existing_row[self.begin_date_column]
                excludes = set(self.primary_key)
                excludes.add(self.end_date_column)
                upsert_row = self.prior_upsert_row.subset(exclude=excludes)
                upsert_row.set_keeping_parent(self.begin_date_column, existing_begin_date)
                super().upsert(
                    source_row=upsert_row,
                    lookup_name=self.prior_upsert_lookup_name,
                    update_callback=self.prior_upsert_update_callback,
                    insert_callback=self.prior_upsert_insert_callback,
                    stat_name=self.prior_upsert_stat_name,
                    parent_stats=self.prior_upsert_parent_stats,
                )
            else:
                pass
        else:
            pass

    def upsert(self,
               source_row: typing.Union[Row, typing.List[Row]],
               lookup_name: str = None,
               skip_update_check_on: list = None,
               do_not_update: list = None,
               additional_update_values: dict = None,
               additional_insert_values: dict = None,
               update_callback: typing.Callable[[list, Row], None] = None,
               insert_callback: typing.Callable[[Row], None] = None,
               source_excludes: frozenset = None,
               target_excludes: frozenset = None,
               stat_name: str = 'upsert',
               parent_stats: Statistics = None,
               **kwargs
               ):
        """
        Update (if changed) or Insert a row in the table.
        Returns the row found/inserted, with the auto-generated key (if that feature is enabled)

        Parameters
        ----------
        source_row: :class:`~bi_etl.components.row.row_case_insensitive.Row`
            Row to upsert
        lookup_name: str
            The name of the lookup (see :meth:`define_lookup`) to use when searching for an existing row.
        skip_update_check_on: list
            List of column names to not compare old vs new for updates.
        do_not_update: list
             List of columns to never update.
        additional_update_values: dict
            Additional updates to apply when updating
        additional_insert_values: dict
            Additional values to set on each row when inserting.
        update_callback: func
            Function to pass updated rows to. Function should not modify row.
        insert_callback: func
            Function to pass inserted rows to. Function should not modify row.
        source_excludes:
            list of source columns to exclude when mapping to this Table.
        target_excludes:
            list of Table columns to exclude when mapping from the source row
        stat_name: string
            Name of this step for the ETLTask statistics. Default = 'upsert'
        parent_stats: bi_etl.statistics.Statistics
            Optional Statistics object to nest this steps statistics in.
            Default is to place statistics in the ETLTask level statistics.
        kwargs:
            effective_date: datetime
                The effective date to use for the update
        """
        stats = self.get_stats_entry(stat_name, parent_stats=parent_stats)
        stats.timer.start()
        stats['upsert source row count'] += 1

        self._set_upsert_mode()

        self.begin()

        effective_date = kwargs.get('effective_date')
        if effective_date is None:
            if self.begin_date_column in source_row and source_row[self.begin_date_column] is not None:
                pass
            else:
                source_row.set_keeping_parent(self.begin_date_column, self.default_effective_date)
        else:
            date_coerce = self.get_coerce_method(self.begin_date_column)
            source_row.set_keeping_parent(self.begin_date_column, date_coerce(effective_date))

        target_excludes = self._target_excludes_for_updates(target_excludes)

        source_mapped_as_target_row = self.build_row(source_row=source_row,
                                                     source_excludes=source_excludes,
                                                     target_excludes=target_excludes,
                                                     parent_stats=stats,
                                                     )

        if self.track_source_rows:
            # Keep track of source records so we can check if target rows don't exist in source
            self.source_keys_processed.add(self.get_natural_key_tuple(source_mapped_as_target_row))

        if lookup_name is None:
            lookup = self.get_default_lookup(source_row.iteration_header)
            lookup_name = lookup.lookup_name
        else:
            lookup = self.get_lookup(lookup_name)

        if len(lookup.lookup_keys) == 1:  # Primary key, it can't be anything else unless not valid.
            # PK lookup doesn't need to deal with list of dates, so call the parent upsert routine directly
            return super().upsert(
                source_row=source_mapped_as_target_row,
                lookup_name=lookup_name,
                skip_update_check_on=skip_update_check_on,
                do_not_update=do_not_update,
                additional_update_values=additional_update_values,
                additional_insert_values=additional_insert_values,
                update_callback=update_callback,
                insert_callback=insert_callback,
                source_excludes=source_excludes,
                target_excludes=target_excludes,
                stat_name=stat_name,
                parent_stats=parent_stats,
                )
        else:
            return self._upsert_version_set(
                source_mapped_as_target_row=source_mapped_as_target_row,
                lookup_name=lookup_name,
                skip_update_check_on=skip_update_check_on,
                do_not_update=do_not_update,
                additional_update_values=additional_update_values,
                additional_insert_values=additional_insert_values,
                update_callback=update_callback,
                insert_callback=insert_callback,
                source_excludes=source_excludes,
                target_excludes=target_excludes,
                stat_name=stat_name,
                parent_stats=parent_stats,
                )

    def update_not_in_set(
            self,
            updates_to_make: typing.Union[dict, Row],
            set_of_key_tuples: set,
            lookup_name: str = None,
            criteria_list: list = None,
            criteria_dict: dict = None,
            use_cache_as_source: bool = True,
            progress_frequency: int = None,
            stat_name: str = 'update_not_in_set',
            parent_stats: Statistics = None,
            **kwargs
    ):
        """
        Applies update to rows matching criteria that are not in the list_of_key_tuples pass in.

        Parameters
        ----------
        updates_to_make: :class:`Row`
            Row or dict of updates to make
        set_of_key_tuples: set
            Set of tuples comprising the primary key values. This list represents the rows that should *not* be updated.
        lookup_name: str
            The name of the lookup (see :meth:`define_lookup`) to use when searching for an existing row.
        criteria_list : string or list of strings
            Each string value will be passed to :meth:`sqlalchemy.sql.expression.Select.where`.
            https://goo.gl/JlY9us
        criteria_dict : dict
            Dict keys should be columns, values are set using = or in
        use_cache_as_source: bool
            Attempt to read existing rows from the cache?
        stat_name: string
            Name of this step for the ETLTask statistics. Default = 'delete_not_in_set'
        progress_frequency: int
            How often (in seconds) to output progress messages. Default = 10.  None for no progress messages.
        parent_stats: bi_etl.statistics.Statistics
            Optional Statistics object to nest this steps statistics in.
            Default is to place statistics in the ETLTask level statistics.
        kwargs:
            effective_date: datetime
                The effective date to use for the update
        """
        if criteria_dict is None:
            criteria_dict = {}

        criteria_dict[self.delete_flag] = self.delete_flag_no
        stats = self.get_unique_stats_entry(stat_name, parent_stats=parent_stats)
        stats['rows read'] = 0
        stats['updates count'] = 0

        stats.timer.start()
        self.begin()

        if progress_frequency is None:
            progress_frequency = self.progress_frequency

        progress_timer = Timer()

        # Turn off read progress reports
        saved_progress_frequency = self.progress_frequency
        self.progress_frequency = None

        if lookup_name is None:
            lookup = self.get_nk_lookup()
        else:
            lookup = self.get_lookup(lookup_name)

        # Ensure that PK columns are returned in addition to whatever lookup is used
        keys = list(lookup.lookup_keys)
        for pk_col in self.primary_key:
            if pk_col not in keys:
                keys.append(pk_col)

        # Note, here we select only lookup + PK columns from self
        row_iter = self.where(column_list=keys,
                              criteria_list=criteria_list,
                              criteria_dict=criteria_dict,
                              use_cache_as_source=use_cache_as_source,
                              parent_stats=stats)
        # Save the iterator as a list since we'll be modifying the cache as we run through it
        rows_list = list(row_iter)
        for row in rows_list:
            pk_val = row[self.primary_key[0]]
            if not isinstance(pk_val, int):
                raise ValueError(f"Row has {pk_val} (not int) primary key value for {self.primary_key[0]} in {row.as_dict}")
            elif pk_val < 0:
                # Skip special rows
                continue

            if row.status == RowStatus.unknown:
                pass
            stats['rows read'] += 1
            existing_key = lookup.get_hashable_combined_key(row)

            if 0 < progress_frequency <= progress_timer.seconds_elapsed:
                progress_timer.reset()
                self.log.info(f"update_not_in_set current current row#={stats['rows read']:,} row key={existing_key} updates done so far = {stats['updates count']:,}")

            if existing_key not in set_of_key_tuples:
                stats['updates count'] += 1

                # Direct update method. This will be slow due to new statement per update
                # and also update statements per row
                # self.update(updates_to_make, key_names=lookup.lookup_keys, key_values=existing_key, connection_name='update')

                try:
                    # First we need the entire existing row (all versions)
                    target_row_list = lookup.find_versions_list(row)
                except NoResultFound:
                    raise RuntimeError(
                        f"keys {row.as_key_value_list} found in database or cache but not found now by get_by_lookup"
                    )

                for target_row in target_row_list:
                    # Then we can apply the updates to it using Table non-versioned updates
                    Table.apply_updates(
                        self=self,
                        row=target_row,
                        additional_update_values=updates_to_make,
                        allow_insert=False,
                        parent_stats=stats,
                    )

        stats.timer.stop()

        # Restore saved progress_frequency
        self.progress_frequency = saved_progress_frequency

    def commit(self,
               stat_name='commit',
               parent_stats=None,
               print_to_log=True,
               connection_name: str = None,
               begin_new: bool = True,
               ):
        if len(self.prior_upsert_existing_rows_list) > 0:
            # End the final batch
            self.finish_pending_existing_rows()
        super().commit(stat_name=stat_name,
                       parent_stats=parent_stats,
                       print_to_log=print_to_log,
                       connection_name=connection_name,
                       begin_new=begin_new,
                       )

    def sql_upsert(
            self,
            source_table: ReadOnlyTable,
            source_excludes: typing.Optional[frozenset] = None,
            target_excludes: typing.Optional[frozenset] = None,
            skip_update_check_on: typing.Optional[frozenset] = None,
            connection_name: str = 'sql_upsert',
            stat_name: str = 'upsert_db_exclusive',
            parent_stats: typing.Optional[Statistics] = None,
            **kwargs
    ):
        upsert_db_excl_stats = self.get_stats_entry(stat_name, parent_stats=parent_stats)
        upsert_db_excl_stats.print_start_stop_times = False
        upsert_db_excl_stats.timer.start()
        upsert_db_excl_stats['calls'] += 1

        source_row_columns_set = source_table.column_names_set
        if source_excludes is not None:
            source_row_columns_set = source_row_columns_set - source_excludes

        target_column_set = self.column_names_set
        if target_excludes is not None:
            target_column_set = target_column_set - target_excludes

        matching_columns = {column_name for column_name in source_row_columns_set if column_name in target_column_set}

        compare_columns = matching_columns - {self.begin_date_column, self.end_date_column}

        non_nk_columns = matching_columns - set(self.natural_key)

        if self.last_update_date is not None:
            compare_columns -= {self.last_update_date}

        if len(self.primary_key) != 1:
            raise ValueError(f"{self}.upsert_db_exclusive requires single primary_key column. Got {self.primary_key}")
        primary_key_name = self.primary_key[0]

        now_str = self.get_current_time().isoformat()

        # TODO: Need to "deduplicate" source versions if two have identical values except for the effective_date_column
        deduplicated_source_table = 'public.deduplicated_source'
        self.database.drop_table_if_exists(deduplicated_source_table, connection_name=connection_name, auto_close=False, transaction=False)


        deduplicate_source_sql = f"""
               CREATE TABLE {deduplicated_source_table} AS
               WITH source_next_prev AS (
                   SELECT
                       {Table._sql_indent_join([col for col in matching_columns], 23, suffix=',')},
                       {kwargs['source_begin_date_column']},
                       LAG({kwargs['source_begin_date_column']}) OVER (
                            PARTITION BY 
                                {Table._sql_indent_join([nk_col for nk_col in self.natural_key], 32, suffix=',')} 
                            ORDER BY {kwargs['source_begin_date_column']}
                        ) as prev_effective_date
                        /*
                        LEAD({kwargs['source_begin_date_column']}) OVER (
                            PARTITION BY 
                                {Table._sql_indent_join([nk_col for nk_col in self.natural_key], 32, suffix=',')} 
                            ORDER BY {kwargs['source_begin_date_column']}
                        ) as next_effective_date
                        */
                   FROM {source_table.qualified_table_name} s
               )
               , changes_only AS (  
               SELECT r.*
               FROM source_next_prev r
                    LEFT OUTER JOIN 
                    source_next_prev prev
                       ON {Table._sql_indent_join([f"prev.{nk_col} = r.{nk_col}" for nk_col in self.natural_key], 32, prefix='AND ')}
                       AND prev.{kwargs['source_begin_date_column']} = r.prev_effective_date
                WHERE 
                    {Table._sql_indent_join([Table._sql_column_not_equals(m_col, 'prev', 'r') for m_col in non_nk_columns], 20, prefix='OR ')}
               )    
               SELECT 
                    changes_only.*,
                    /* Note we can't use the next_effective_date since we might have skipped one with no changes */
                    COALESCE(
                         LEAD({kwargs['source_begin_date_column']}) OVER (
                            PARTITION BY 
                                {Table._sql_indent_join([nk_col for nk_col in self.natural_key], 32, suffix=',')} 
                            ORDER BY {kwargs['source_begin_date_column']}
                        ), 
                        {self._sql_date_literal(self.default_end_date)}
                    ) as end_effective_date
               FROM changes_only                 
               """
        self.log.debug('-' * 80)
        self.log.debug('deduplicate_source_sql')
        self.log.debug('-' * 80)
        sql_timer = Timer()
        self.log.debug(deduplicate_source_sql)
        self.execute(deduplicate_source_sql, connection_name=connection_name)
        self.log.debug(f"Execution time ={sql_timer.seconds_elapsed_formatted}")

        # Insert rows for new Natural Keys
        self._sql_insert_new(
            source_table=deduplicated_source_table,
            matching_columns=matching_columns,
            effective_date_column=kwargs['source_begin_date_column'],
            connection_name=connection_name,
        )

        # unified_dates_table = '#unified_dates_table'
        unified_dates_table = 'public.derek_unified_dates_table'
        self.database.drop_table_if_exists(unified_dates_table, connection_name=connection_name, auto_close=False, transaction=False)

        unified_dates_sql = f"""
            CREATE TABLE {unified_dates_table} AS
            WITH pass1 as (
                SELECT DISTINCT 
                    {Table._sql_indent_join([f"s.{nk_col}" for nk_col in self.natural_key], 16, suffix=',')},
                    s.{kwargs['source_begin_date_column']} as effective_date
                FROM {deduplicated_source_table} s
                UNION
                SELECT
                    {Table._sql_indent_join([f"e.{nk_col}" for nk_col in self.natural_key], 16, suffix=',')},
                    e.{self.begin_date_column} as effective_date
                FROM {self.qualified_table_name} e
            )
            SELECT 
            pass1.*,
            COALESCE(
                LEAD({self._sql_add_timedelta('pass1.effective_date', self._end_date_timedelta)}) OVER (
                    PARTITION BY {Table._sql_indent_join([nk_col for nk_col in self.natural_key], 16, suffix=',')}
                    ORDER BY effective_date
                    ),
                {self._sql_date_literal(self.default_end_date)}
            ) as new_end_effective_date
           FROM pass1 
        """
        self.log.debug('-' * 80)
        self.log.debug('unified_dates_sql')
        self.log.debug('-' * 80)
        sql_timer = Timer()
        self.log.debug(unified_dates_sql)
        self.execute(unified_dates_sql, connection_name=connection_name)
        self.log.debug(f"Execution time ={sql_timer.seconds_elapsed_formatted}")

        # TODO: Need DB specific CREATE TABLE AS or SELECT ... INTO
        # derek_updated_row_nk_table = '#updated_row_nk'
        updated_row_nk_table = 'public.derek_updated_row_nk'

        self.database.drop_table_if_exists(updated_row_nk_table, connection_name=connection_name, auto_close=False, transaction=False)

        find_updates_sql = f"""
            CREATE TABLE {updated_row_nk_table} AS
            SELECT 
                {Table._sql_indent_join([f"e.{nk_col}" for nk_col in self.natural_key], 16, suffix=',')},
                s.{kwargs['source_begin_date_column']} as source_effective_date,
                e.{self.begin_date_column} as target_begin_date,
                e.{self.end_date_column} as target_end_date,
                d.new_end_effective_date,
                {Table._sql_indent_join([f"CASE WHEN {Table._sql_column_not_equals(m_col, 'e', 's')} THEN 1 ELSE 0 END as chg_in_{m_col}" for m_col in non_nk_columns], 16, suffix=',')}
            FROM {unified_dates_table} d
                 INNER JOIN
                 {self.qualified_table_name} e
                    ON {Table._sql_indent_join([f"d.{nk_col} = e.{nk_col}" for nk_col in self.natural_key], 19, prefix='AND ')}
                   AND d.effective_date BETWEEN e.{self.begin_date_column} AND e.{self.end_date_column}
                 INNER JOIN
                 {deduplicated_source_table} s
                    ON {Table._sql_indent_join([f"e.{nk_col} = s.{nk_col}" for nk_col in self.natural_key], 19, prefix='AND ')}
                   AND d.effective_date BETWEEN s.{kwargs['source_begin_date_column']} AND s.{kwargs['source_end_date_column']} 
            WHERE ( 
                    {Table._sql_indent_join([Table._sql_column_not_equals(m_col, 'e', 's') for m_col in non_nk_columns], 20, prefix='OR ')}
                    OR e.{self.delete_flag} != '{self.delete_flag_no}'
                  )
        """
        self.log.debug('-' * 80)
        self.log.debug('find_updates_sql')
        self.log.debug('-' * 80)
        sql_timer = Timer()
        self.log.debug(find_updates_sql)
        results = self.execute(find_updates_sql, connection_name=connection_name)
        self.log.debug(f"Execution time ={sql_timer.seconds_elapsed_formatted}")

        # Note: This SQL will only work if generated AFTER the insert_new_sql has been run on the DB
        # otherwise the primary_key values will conflict.

        # TODO type 1 srgt
        insert_updates_sql = f"""                
            INSERT INTO {self.qualified_table_name} (
                {primary_key_name},
                {self.delete_flag},
                {self.begin_date_column},
                {self.end_date_column},
                {self.last_update_date},
                {Table._sql_indent_join(matching_columns, 16, suffix=',')}
            )  
            WITH max_srgt AS (SELECT coalesce(max({primary_key_name}),0)  as max_srgt FROM {self.qualified_table_name})
            SELECT 
                max_srgt.max_srgt + ROW_NUMBER() OVER (order by 1) as {primary_key_name},
                '{self.delete_flag_no}' as {self.delete_flag},
                u.source_effective_date as {self.begin_date_column},
                u.new_end_effective_date as {self.end_date_column},
                '{now_str}' as {self.last_update_date},
                {Table._sql_indent_join([f's.{col}' for col in matching_columns], 16, suffix=',')}
            FROM max_srgt
                 CROSS JOIN 
                 {deduplicated_source_table} s
                 INNER JOIN
                 {updated_row_nk_table} u
                   ON {Table._sql_indent_join([f"u.{nk_col} = s.{nk_col}" for nk_col in self.natural_key], 18, prefix='AND')}
                  AND s.{kwargs['source_begin_date_column']} = u.source_effective_date
            /* Do not insert entries for when dates match exactly, those must be updated in place */
            WHERE u.source_effective_date != u.target_begin_date
        """

        self.log.debug('-' * 80)
        self.log.debug('insert_updates_sql')
        self.log.debug('-' * 80)
        self.log.debug(insert_updates_sql)
        sql_timer = Timer()
        results = self.execute(insert_updates_sql, connection_name=connection_name)
        self.log.debug(f"Impacted rows = {results.rowcount:,} (meaningful count? {results.supports_sane_rowcount()})")
        self.log.debug(f"Execution time ={sql_timer.seconds_elapsed_formatted}")

        joins = [(nk_col, ("u", nk_col)) for nk_col in self.natural_key]
        joins.append((self.begin_date_column, ('u', 'target_begin_date')))
        self._sql_update_from(
            update_name='update_retire',
            source_sql=f"{updated_row_nk_table} u",
            list_of_joins=joins,
            # Do not insert entries for when dates match exactly, those must be updated in place.
            extra_where="u.source_effective_date != u.target_begin_date",
            target_alias_in_source='e',
            list_of_sets=[
                (
                    self.end_date_column,
                    f"u.new_end_effective_date"
                ),
            ]
        )

        self._sql_update_from(
            update_name='update_in_place',
            source_sql=f"""
                {updated_row_nk_table} u
                 INNER JOIN
                 {deduplicated_source_table} s
                   ON {Table._sql_indent_join([f"s.{nk_col} = u.{nk_col}" for nk_col in self.natural_key], 18, prefix='AND ')}
                  AND s.{kwargs['source_begin_date_column']} = u.source_effective_date
            """,
            list_of_joins=joins,
            target_alias_in_source='e',
            list_of_sets=[
                (m_col,
                 f"s.{m_col}")
                for m_col in non_nk_columns
            ]
        )

        self.log.debug('=' * 80)

        self.commit()

        self.close_connection(connection_name=connection_name)

        upsert_db_excl_stats.timer.stop()




# -*- coding: utf-8 -*-
"""
Created on Dec 23, 2015

@author: Derek Wood
"""
import logging
import textwrap

import sqlalchemy
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.orm import Session
from sqlalchemy.sql.schema import DEFAULT_NAMING_CONVENTION

from bi_etl.utility.case_insentive_set import CaseInsentiveSet


class DatabaseMetadata(sqlalchemy.schema.MetaData):
    """
    A light wrapper over :class:`sqlalchemy.schema.MetaData`
    """

    def __init__(self,
                 bind=None,
                 reflect=False,
                 schema=None,
                 quote_schema=None,
                 naming_convention=DEFAULT_NAMING_CONVENTION,
                 info=None,
                 database_name=None,
                 uses_bytes_length_limits=None,
                 ):
        super().__init__(
            bind=bind,
            schema=schema,
            quote_schema=quote_schema,
            naming_convention=naming_convention,
            info=info,
        )
        # Save parameters not saved by the base class for use in __reduce_ex__

        self._save_reflect = reflect
        self._save_quote_schema = quote_schema

        self._table_inventory = None
        self.database_name = database_name
        self._uses_bytes_length_limits = uses_bytes_length_limits

        self._connection_pool = dict()
        self._transactions = dict()
        self.default_connection_name = 'default'
        self._autocommit_engine = self.bind.execution_options(isolation_level="AUTOCOMMIT")

        self.log = logging.getLogger(f"{self.__class__.__module__}.{self.__class__.__name__}")

    def __reduce_ex__(self, protocol):
        return (
            # A callable object that will be called to create the initial version of the object.
            self.__class__,

            # A tuple of arguments for the callable object. An empty tuple must be given if the callable does not accept any argument
            (self.bind.url, self._save_reflect, self.schema, self._save_quote_schema, self.naming_convention, self.info, self.database_name, self._uses_bytes_length_limits),

            # Optionally, the object’s state, which will be passed to the object’s __setstate__() method as previously described.
            # If the object has no such method then, the value must be a dictionary and it will be added to the object’s __dict__ attribute.
            None,

            # Optionally, an iterator (and not a sequence) yielding successive items.
            # These items will be appended to the object either using obj.append(item) or, in batch, using obj.extend(list_of_items).

            # Optionally, an iterator (not a sequence) yielding successive key-value pairs.
            # These items will be stored to the object using obj[key] = value

            # PROTOCOL 5+ only
            # Optionally, a callable with a (obj, state) signature.
            # This callable allows the user to programmatically control the state-updating behavior of a specific object,
            # instead of using obj’s static __setstate__() method.
            # If not None, this callable will have priority over obj’s __setstate__().
        )

    def _set_parent(self, parent):
        pass

    def resolve_connection_name(self, connection_name: str = None) -> str:
        if connection_name is None:
            connection_name = self.default_connection_name
        # When using sqlite don't make new connections, reuse the existing one
        if self.dialect_name == 'sqlite':
            connection_name = 'sqlite'
        return connection_name

    def _connect(self, auto_commit: bool = False) -> sqlalchemy.engine.base.Connection:
        if auto_commit:
            self.log.debug(f"Connecting autocommit connection {self._autocommit_engine}")
            return self._autocommit_engine.connect()
        else:
            self.log.debug(f"Connecting connection {self.bind}")
            return self.bind.connect()

    def connection(
            self,
            connection_name: str = None,
            open_if_not_exist: bool = True,
            open_if_closed: bool = True,
            auto_commit: bool = False,
    ) -> sqlalchemy.engine.base.Connection:
        connection_name = self.resolve_connection_name(connection_name)
        connection_key = (auto_commit, connection_name,)
        if connection_key in self._connection_pool:
            con = self._connection_pool[connection_key]
            if con.closed and open_if_closed:
                con = self._connect(auto_commit=auto_commit)
                self._connection_pool[connection_key] = con
        else:
            if open_if_not_exist:
                con = self._connect(auto_commit=auto_commit)
                self._connection_pool[connection_key] = con
            else:
                raise ValueError(f"Connection {connection_name} does not exist, and open_if_not_exist = False")
        return con

    def connect(
            self,
            connection_name: str = None,
            auto_commit: bool = False,
    ) -> sqlalchemy.engine.base.Connection:
        return self.connection(
            connection_name,
            open_if_not_exist=True,
            open_if_closed=True,
            auto_commit=auto_commit,
        )

    def is_connected(self, connection_name: str = None) -> bool:
        try:
            con = self.connection(connection_name, open_if_not_exist=False, open_if_closed=False)
            return con.closed
        except ValueError:
            return False

    def close_connection(self, connection_name: str = None):
        try:
            con = self.connection(connection_name, open_if_not_exist=False, open_if_closed=False)
            con.close()
        except ValueError:
            pass

    def close_connections(self, exceptions: set = None):
        if exceptions is None:
            exceptions = set()
        for connection_key, con in self._connection_pool.items():
            auto_commit, connection_name = connection_key
            if connection_name not in exceptions:
                self.log.debug(f'Closing connection {self} {connection_name}')
                con.close()

    def dispose(self):
        """
        This method leaves the possibility of checked-out connections
        remaining open, as it only affects connections that are
        idle in the pool.
        """
        self.close_connections()
        self.bind.pool.dispose()

    def session(self, auto_commit: bool = False):
        return Session(bind=self.bind, autocommit=auto_commit)

    def _begin(self, connection_name: str) -> sqlalchemy.engine.base.Transaction:
        tx = self.connection(connection_name=connection_name).begin()
        self._transactions[connection_name] = tx
        return tx

    def begin(self, connection_name: str = None) -> sqlalchemy.engine.base.Transaction:
        connection_name = self.resolve_connection_name(connection_name)
        if connection_name not in self._transactions:
            tx = self._begin(connection_name)
        else:
            tx = self._transactions[connection_name]
            if not tx.is_active:
                tx = self._begin(connection_name)
        return tx

    def has_active_transaction(self, connection_name: str = None):
        connection_name = self.resolve_connection_name(connection_name)
        if connection_name not in self._transactions:
            return False
        else:
            tx = self._transactions[connection_name]
            return tx.is_active

    def commit(self, connection_name: str = None):
        """
        Commit based on a connection name rather than via a
        'sqlalchemy.engine.base.Transaction' object (which you could call .commit() on

        Parameters
        ----------
        connection_name
        """
        connection_name = self.resolve_connection_name(connection_name)
        if connection_name not in self._transactions:
            self.log.debug(f"Commit: There is no transaction recorded for {self} {connection_name}")
        else:
            tx = self._transactions[connection_name]
            if tx.is_active:
                self.log.info(f'Commit on {self} {connection_name} connection started')
                tx.commit()
                self.log.debug(f'Commit on {self} {connection_name} connection done')
            else:
                self.log.info(f'Connection {self} {connection_name} transaction not active (commit called)')

    def rollback(self, connection_name: str = None):
        connection_name = self.resolve_connection_name(connection_name)
        if connection_name not in self._transactions:
            raise RuntimeError(f"rollback: There is no transaction recorded for {self} {connection_name}")
        else:
            tx = self._transactions[connection_name]
            if tx.is_active:
                tx.rollback()
                self.log.info(f'Rollback on {self} {connection_name} connection done')
            else:
                raise RuntimeError(f'Connection {self} {connection_name} transaction not active (rollback called)')

    def execute(
            self,
            sql,
            *list_params,
            transaction: bool = True,
            auto_close: bool = True,
            auto_commit: bool = False,
            connection_name: str = None,
            **params
    ):
        connection = None
        transaction_ctl = None
        try:
            connection = self.connect(
                connection_name=connection_name,
                auto_commit=auto_commit,
            )
            if transaction:
                transaction_ctl = connection.begin()
            else:
                transaction_ctl = None
            result = connection.execute(sql, *list_params, **params)
            return result
        finally:
            if transaction:
                if transaction_ctl is not None:
                    transaction_ctl.commit()
            if auto_close:
                if connection is not None:
                    connection.close()

    def execute_procedure(
            self,
            procedure_name,
            *args,
            return_results=False,
            dpapi_connection=None
    ):
        """
        Execute a stored procedure 
        
        Parameters
        ----------
        procedure_name: str
            The procedure to run.
        args:
            The arguments to pass

        return_results:
            Needs to be a keyword param. Should we try and get result rows
            from the procedure.

        dpapi_connection:
            A raw dpapi connection to use. Optional.
            
        Raises
        ------
        sqlalchemy.exc.DBAPIError:
            API error            
        sqlalchemy.exc.DatabaseError:
            Proxy for database error
        """
        log = logging.getLogger(__name__)
        log.debug("Calling procedure {} {}".format(procedure_name, args))

        if dpapi_connection is None:
            dpapi_connection = self.bind.raw_connection()
            close_connection = True
        else:
            close_connection = False
        results = None
        try:
            cursor = dpapi_connection.cursor()
            if hasattr(cursor, 'callproc'):
                cursor.callproc(procedure_name, args)
                if return_results:
                    results = list(cursor.fetchall())
                cursor.close()
            else:
                # Stopped using CALL because of issues like those mentioned on https://stackoverflow.com/a/34179375
                # if False: # 'pyodbc' in self.bind.dialect.dialect_description == 'mssql+pyodbc':
                #     if len(args) > 0:
                #         sql = f"{{CALL {procedure_name}({','.join([qmark for qmark in ['?'] * len(args)])}) }}"
                #     else:
                #         sql = f"{{CALL {procedure_name}}}"
                # else:
                # sql = f"EXEC {procedure_name} {','.join([qmark for qmark in ['?'] * len(args)])}"
                sql = f"EXEC {procedure_name} "
                args2 = []
                delim = ''
                for arg in args:
                    if isinstance(arg, str):
                        arg = arg.strip()
                    # Handle keyword named parameters
                    if arg[0] == '@':
                        param, value = arg.split('=')
                        param = param.strip()
                        param = f'{param}=?'
                        value = value.strip()
                        # Likely the opening quote of the value has not been removed yet
                        if value[0] == "'":
                            value = value[1:]
                    else:
                        param = '?'
                        value = arg
                    sql += delim + param
                    delim = ', '
                    args2.append(value)

                cursor.execute(sql, args2)

                if return_results:
                    results = list(cursor.fetchall())
                cursor.close()
            dpapi_connection.commit()
        finally:
            if close_connection:
                dpapi_connection.close()
        return results

    def execute_direct(
            self,
            sql,
            return_results=False
    ):
        log = logging.getLogger(__name__)
        log.debug(sql)
        dpapi_connection = self.bind.raw_connection()
        try:
            cursor = dpapi_connection.cursor()
            cursor.execute(sql)
            results = None
            if return_results:
                results = list(cursor.fetchall())
            cursor.close()
            dpapi_connection.commit()
        finally:
            dpapi_connection.close()
        return results

    def table_inventory(self, schema=None, force_reload=False):
        if self._table_inventory is None:
            self._table_inventory = dict()
        if schema not in self._table_inventory or force_reload:
            try:
                from sqlalchemy import inspect
            except ImportError:
                inspect = Inspector.from_engine
            inspector = inspect(self.bind)
            self._table_inventory[schema] = CaseInsentiveSet(inspector.get_table_names(schema=schema))
        return self._table_inventory[schema]

    @staticmethod
    def qualified_name(schema, table):
        if schema is not None:
            return schema + '.' + table
        else:
            return table

    def rename_table(self, schema, table_name, new_table_name):
        if self.dialect_name == 'mssql':
            self.execute_procedure(
                'sp_rename',
                self.qualified_name(schema, table_name),
                new_table_name
            )
        else:
            sql = f"alter table {self.qualified_name(schema, table_name)} rename to {new_table_name}"
            self.log.debug(sql)
            self.execute(sql)

    def drop_table_if_exists(
            self,
            table_name,
            schema=None,
            connection_name: str = None,
            transaction: bool = False,
            auto_close: bool = False,
    ):
        if schema is None:
            if '.' in table_name:
                schema, table_name = table_name.split('.')
        # SQL Server 2016+ can use IF EXISTS but rather than checking version use compatible mode
        if self.dialect_name == 'mssql':
            if table_name[0] == '#':
                # Temp table
                sql = textwrap.dedent(f"""\
                    IF OBJECT_ID('tempdb.dbo.{table_name}', 'U') IS NOT NULL 
                    DROP TABLE {self.qualified_name(schema, table_name)}; 
                """)
            else:
                sql = textwrap.dedent(f"""\
                    IF OBJECT_ID('{self.qualified_name(schema, table_name)}', 'U') IS NOT NULL 
                    DROP TABLE {self.qualified_name(schema, table_name)}; 
                """)
        elif self.dialect_name == 'oracle':
            sql = textwrap.dedent(f"""\
            BEGIN
               EXECUTE IMMEDIATE 'DROP TABLE {table_name}';
            EXCEPTION
               WHEN OTHERS THEN
                  IF SQLCODE != -942 THEN
                     RAISE;
                  END IF;
            END;
            """)
        else:
            sql = f"drop table IF EXISTS {self.qualified_name(schema, table_name)}"
        self.log.debug(sql)
        self.execute(
            sql,
            transaction=transaction,
            auto_close=auto_close,
            connection_name=connection_name,
        )

    @property
    def dialect(self):
        return self.bind.dialect

    @property
    def dialect_name(self):
        return self.bind.dialect.name

    @property
    def uses_bytes_length_limits(self):
        if self._uses_bytes_length_limits is None:
            # Note: Oracle can use either VARCHAR2(10 CHAR) or VARCHAR2(10 BYTE)
            #       However, if not specified (and NLS_LENGTH_SEMANTICS is default), it's char so we assume that.
            if self.dialect_name in {
                'redshift', 'oracle'
            }:
                self._uses_bytes_length_limits = True
            else:
                self._uses_bytes_length_limits = False
        return self._uses_bytes_length_limits



from bi_etl.components.get_next_key.base_table_memory import BaseTableMemory


class SharedTableMemory(BaseTableMemory):
    def __init__(self, manager, chunk_size: int):
        # Table will need to be set on local process before use
        super().__init__(table=None)
        self.chunk_size = chunk_size
        self.lock = manager.Lock()
        self.current_key_values = manager.dict()
        # These need to be None on init because init is done on main process
        self._local_current_keys = None
        self._local_max_allocated_keys = None
        self.log.debug(f"SharedTableMemory chunk_size= {chunk_size}")

    def get_key_from_shared(self, column: str) -> int:
        column_obj = self.table.get_column(column)
        column = column_obj.name
        with self.lock:
            current_max = self.current_key_values.get(column)
            if current_max is None:
                next_key = self.get_next_from_database(column_obj)
            else:
                next_key = current_max + 1
            local_max_allocated = next_key + self.chunk_size
            self.current_key_values[column] = local_max_allocated
            self._local_max_allocated_keys[column] = local_max_allocated
        return next_key

    def get_next_key(self, column: str) -> int:
        column_obj = self.table.get_column(column)
        column = column_obj.name
        if self._local_current_keys is None:
            self._local_current_keys = dict()
            self._local_max_allocated_keys = dict()
        current_max = self._local_current_keys.get(column)
        if current_max is None:
            next_key = self.get_key_from_shared(column)
        else:
            local_max_allocated = self._local_max_allocated_keys.get(column)
            if current_max >= local_max_allocated:
                # defer to parent to get max
                next_key = self.get_key_from_shared(column)
            else:
                next_key = current_max + 1
        self._local_current_keys[column] = next_key
        return next_key



import logging

from sqlalchemy import Column


class BaseTableMemory(object):
    def __init__(self, table):
        self.table = table
        self.log = logging.getLogger(f"{self.__class__.__name__}.{table}")

    def get_next_from_database(self, column_obj: Column) -> int:
        current_max = self.table.max(column_obj)
        # In case the database call returns None
        if current_max is None:
            next_key = 0
        else:
            # Make sure max is an integer
            next_key = int(current_max)
        # Check for negative (special values) set max to 0
        if next_key <= 0:
            next_key = 1
            msg = "(and not negative value)"
        else:
            msg = ''
            next_key = current_max + 1
        self.log.info(f"Initialize sequence for {column_obj} with value {next_key} {msg}")
        return next_key



"""
Created on March 28, 2022

@author: 
"""
from typing import *

from bi_etl.boto3_helper.session import Boto3_Base


class DynamoDB(Boto3_Base):
    SERVICE = 'dynamodb'

    def get_dynamo_table(
            self,
            dynamo_table_name,
            region_name: str = None
    ):
        if region_name is not None:
            self.region_name = region_name
        dynamodb = self.resource

        return dynamodb.Table(dynamo_table_name)

    @staticmethod
    def query_dynamo_table(dynamo_table, scan_args_list: Iterator[dict]) -> Iterator[dict]:
        data = []
        for scan_args in scan_args_list:
            tbl_data = dynamo_table.query(**scan_args)
            data.extend(tbl_data['Items'])

            while 'LastEvaluatedKey' in tbl_data:
                tbl_data = dynamo_table.query(ExclusiveStartKey=tbl_data['LastEvaluatedKey'], **scan_args)
                data.extend(tbl_data['Items'])

        return data

    def query_dynamo_table_by_name(
            self,
            dynamo_table_name,
            region_name,
            scan_args_list: Iterator[dict]
    ) -> Iterator[dict]:
        dynamo_table = self.get_dynamo_table(dynamo_table_name, region_name=region_name)
        return self.query_dynamo_table(dynamo_table, scan_args_list)

    def scan_dynamo_table_by_name(self, dynamo_table_name, region_name):
        dynamo_table = self.get_dynamo_table(dynamo_table_name, region_name=region_name)
        return self.scan_dynamo_table(dynamo_table)

    def scan_dynamo_table(self, dynamo_table):
        data = []
        tbl_data = dynamo_table.scan()
        data.extend(tbl_data['Items'])

        return data


from pathlib import Path
from typing import Union

import psycopg2
from sqlalchemy.engine import Engine
from sqlalchemy.exc import SQLAlchemyError, DBAPIError


def get_conn(
        dbname,
        username,
        password,
        encoding='UTF-8',
):
    conn = psycopg2.connect(database=dbname, user=username, password=password)
    conn.set_client_encoding(encoding)
    return conn


def get_cursor(
        dbname,
        username,
        password,
        encoding='UTF-8',
):
    conn = get_conn(
        dbname=dbname,
        username=username,
        password=password,
        encoding=encoding
    )
    cur = conn.cursor()
    return cur


def psycopg2_extract_using_engine(
        engine: Engine,
        table_or_query: str,
        output_file_path: Union[str, Path],
        delimiter: str = '|',
        csv_mode: bool = True,
        header: bool = True,
        null: str = '',
        encoding='UTF-8',
):
    """

    Parameters
    ----------
    engine
    table_or_query:
        table_name [ ( column_name [, ...] )
        or (SQL Query)
        For table based, column names are optional.
        For SQL query based, note that parentheses and quotes are required around the query.
    output_file_path
    delimiter
    csv_mode
    header
    null
    encoding

    Returns
    -------

    """
    conn = engine.raw_connection()
    conn.set_client_encoding(encoding)
    cur = conn.cursor()
    if csv_mode:
        if header:
            header_cmd = 'HEADER'
        else:
            header_cmd = ''
        copy_stmt = f"COPY {table_or_query} TO STDOUT WITH CSV {header_cmd} DELIMITER '{delimiter}' NULL '{null}'"
    else:
        copy_stmt = f"COPY {table_or_query} TO STDOUT WITH DELIMITER '{delimiter}' NULL '{null}'"
    with open(output_file_path, 'wt', encoding=encoding, newline='\n') as output_file:
        cur.copy_expert(copy_stmt, output_file)
        # cur.copy_to(output_file, table_or_query, sep=delimiter, null=null)
    conn.close()


def psycopg2_import_using_cursor(
        cursor,
        table_spec: str,
        input_file_path: Union[str, Path],
        block_size: int = 4096,
        delimiter: str = '|',
        csv_mode: bool = True,
        header: bool = True,
        null: str = '',
        encoding='UTF-8',
):
    """

    Parameters
    ----------
    engine
    table_spec:
        table_name [ ( column_name [, ...] )
        Column names are optional unless the columns in the file do not match the order or set of columns
        in the table.
    input_file_path
    block_size
    delimiter
    csv_mode
    header
    null
    encoding
    """
    if csv_mode:
        format_str = 'csv'
    else:
        format_str = 'text'

    if header:
        if csv_mode:
            header_str = 'HEADER'
        else:
            raise ValueError("Header option is allowed only when using CSV format.")
    else:
        header_str = ''

    copy_stmt = f"""
        COPY "{table_spec}" FROM STDIN
        WITH  {format_str}
        DELIMITER '{delimiter}'
        NULL '{null}'
        {header_str}  
        """
    try:
        with open(input_file_path, 'rt', encoding=encoding, newline='\n') as input_file:
            results = cursor.copy_expert(copy_stmt, input_file, size=block_size)
            # if header:
            #     input_file.readline()
            # results = cursor.copy_from(
            #     file=input_file,
            #     table=table_spec,
            #     sep=delimiter,
            #     null=null,
            # )
        return results
    except Exception as e:
        raise DBAPIError(
            statement=copy_stmt,
            params=None,
            orig=e,
        )


def psycopg2_import_using_engine(
        engine: Engine,
        table_spec: str,
        input_file_path: Union[str, Path],
        block_size: int = 4096,
        delimiter: str = '|',
        csv_mode: bool = True,
        header: bool = True,
        null: str = '',
        encoding='UTF-8',
):
    """

    Parameters
    ----------
    engine
    table_spec:
        table_name [ ( column_name [, ...] )
        Column names are optional unless the columns in the file do not match the order or set of columns
        in the table.
    input_file_path
    block_size
    delimiter
    csv_mode
    header
    null
    encoding
    """
    conn = engine.raw_connection()
    conn.set_client_encoding(encoding)
    cursor = conn.cursor()
    results = psycopg2_import_using_cursor(
        cursor=cursor,
        table_spec=table_spec,
        input_file_path=input_file_path,
        block_size=block_size,
        delimiter=delimiter,
        csv_mode=csv_mode,
        header=header,
        null=null,
        encoding=encoding,
    )
    conn.close()
    return results



#!/usr/bin/env python
import argparse
import mmap
import subprocess
from collections import defaultdict
from timeit import default_timer as timer


def mapcount(filename):
    f = open(filename, "r+")
    buf = mmap.mmap(f.fileno(), 0)
    lines = 0
    readline = buf.readline
    while readline():
        lines += 1
    return lines


def simplecount(filename):
    lines = 0
    for _ in open(filename):
        lines += 1
    return lines


# Fastest for small files
def bufcount(filename):
    f = open(filename)
    lines = 0
    buf_size = 1024 * 1024
    read_f = f.read  # loop optimization

    buf = read_f(buf_size)
    while buf:
        lines += buf.count('\n')
        buf = read_f(buf_size)

    return lines


## Fastest for big files
def wccount(filename):
    out = subprocess.Popen(['wc', '-l', filename],
                           stdout=subprocess.PIPE,
                           stderr=subprocess.STDOUT
                           ).communicate()[0]
    return int(out.partition(b' ')[0])


def itercount(filename):
    return sum(1 for _ in open(filename, 'rbU'))


def opcount(fname):
    line_number = 0
    with open(fname) as f:
        for line_number, _ in enumerate(f, 1):
            pass
    return line_number


def kylecount(fname):
    return sum(1 for _ in open(fname))


def countLines(filename):
    return wccount(filename)


def clear_cache():
    """Clear disk cache on Linux."""
    # os.system("sync ; sudo /bin/sh -c 'echo 3 > /proc/sys/vm/drop_caches'")
    pass


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument('---clear-cache', action="store_true")
    parser.add_argument('-n', type=int, default=3)
    parser.add_argument('filename')

    args = parser.parse_args()

    counts = defaultdict()

    if args.clear_cache:
        do_clear_cache = True
    else:
        do_clear_cache = False

    filename = args.filename
    for _ in range(args.n):
        for func in (f
                     for n, f in list(globals().items())
                     if n.endswith('count') and hasattr(f, '__call__')):
            if do_clear_cache:
                clear_cache()
            start_time = timer()
            # http://norvig.com/big.txt
            if filename == 'big.txt':
                assert func(filename) == 128457  # 128457 1095695 6488666 big.txt
            else:
                func(filename)
            counts[func].append(timer() - start_time)

    timings = {}
    for key, vals in list(counts.items()):
        timings[key.__name__] = sum(vals) / float(len(vals)), min(vals)
    width = max(len(n) for n in timings) + 1
    print(("%s %s %s %s" % (
        "function".ljust(width),
        "average, s".rjust(15),
        "min, s".rjust(15),
        "ratio".rjust(15))))
    absmin_ = min(x[1] for x in list(timings.values()))
    for name, (av, min_) in sorted(list(timings.items()), key=lambda x: x[1][1]):
        print(("{} {:15.4f} {:15.4f} {:15.4f}".format(name.ljust(width), av, min_, min_ / absmin_)))


if __name__ == '__main__':
    main()

# function      average, s  min, s  ratio
# wccount            0.005  0.0042   1.00
# bufcount          0.0081  0.0081   1.91
# fadvcount         0.0094  0.0091   2.13
# opcount            0.018   0.015   3.42
# simplecount        0.019   0.016   3.66
# kylecount          0.019   0.017   4.03
# mapcount           0.027   0.021   4.97
# itercount          0.044   0.031   7.21

## python3.1 ginstrom.py
# function      average, s  min, s  ratio
# wccount           0.0049  0.0046   1.00
# itercount          0.021    0.02   4.47
# mapcount           0.023   0.023   5.09
# bufcount           0.034   0.032   7.02
# opcount            0.043   0.043   9.46
# simplecount         0.05   0.046  10.20
# kylecount           0.05    0.05  10.95

## python ginstrom.py /big/mkv/file
# function      average, s  min, s  ratio
# wccount             0.51    0.49   1.00
# opcount              1.8     1.8   3.58
# simplecount          1.8     1.8   3.66
# kylecount            1.9     1.9   3.75
# mapcount              19       2   4.01
# fadvcount            2.3     2.2   4.52
# bufcount             2.3     2.2   4.52
## wc /big/mkv/file
## 7137518   40523351 1836139137 /big/mkv/file

## with --clear-cache
# function      average, s  min, s  ratio
# simplecount         0.06   0.057   1.00
# opcount            0.067   0.057   1.00
# kylecount          0.057   0.057   1.00
# itercount           0.06   0.058   1.02
# mapcount           0.059   0.058   1.02
# fadvcount          0.064   0.058   1.02
# bufcount            0.07   0.062   1.09
# wccount            0.072   0.065   1.15

## python3.1 with --clear-cache
# function      average, s  min, s  ratio
# itercount          0.061   0.057   1.00
# simplecount        0.069   0.061   1.06
# mapcount           0.062   0.061   1.07
# wccount            0.067   0.064   1.11
# kylecount          0.067   0.065   1.12
# opcount            0.072   0.067   1.17
# bufcount           0.083   0.073   1.27



"""
Created on Feb 27, 2015

@author: Derek Wood
"""
# https://www.python.org/dev/peps/pep-0563/
from __future__ import annotations

import typing

from bi_etl.components.row.row import Row
from bi_etl.config.bi_etl_config_base import BI_ETL_Config_Base
from bi_etl.exceptions import NoResultFound
from bi_etl.lookups.lookup import Lookup

if typing.TYPE_CHECKING:
    from bi_etl.components.etlcomponent import ETLComponent

__all__ = ['NonUniqueLookup']


class NonUniqueLookup(Lookup):
    def __init__(self,
                 lookup_name: str,
                 lookup_keys: list,
                 parent_component: ETLComponent,
                 config: BI_ETL_Config_Base = None,
                 use_value_cache: bool = True,
                 value_for_none='<None>',  # Needs to be comparable to datatypes in the key actual None is not.
                 ):
        super(NonUniqueLookup, self).__init__(
            lookup_name=lookup_name,
            lookup_keys=lookup_keys,
            parent_component=parent_component,
            use_value_cache=use_value_cache,
            config=config,
            )
        self.value_for_none = value_for_none
        self._remote_lookup_stmt_no_date = None
        self._len = 0

    def __len__(self):
        return self._len

    def get_list_of_lookup_column_values(self, row: Lookup.ROW_TYPES) -> list:
        results = super().get_list_of_lookup_column_values(row)

        # We expect that non-unique lookups might have nulls, we need to eliminate those
        results_2 = list()
        for value in results:
            if value is None:
                value = self.value_for_none
            results_2.append(value)

        return results_2

    def cache_row(
            self,
            row: Row,
            allow_update: bool = True,
            allow_insert: bool = True,
    ):
        """
        Adds the given row to the cache for this lookup.

        Parameters
        ----------
        row: Row
            The row to cache

        allow_update: boolean
            Allow this method to update an existing row in the cache.

        allow_insert: boolean
            Allow this method to insert a new row into the cache

        Raises
        ------
        ValueError
            If allow_update is False and an already existing row (lookup key) is passed in.

        """
        if self.cache_enabled:
            assert isinstance(row, Row), "cache_row requires Row and not {}".format(type(row))

            if self.use_value_cache:
                self._update_with_value_cache(row)

            lk_tuple = self.get_hashable_combined_key(row)
            if self._cache is None:
                self.init_cache()
            matching_rows = self._cache.get(lk_tuple, None)
            try:
                # noinspection PyUnresolvedReferences
                primary_key = self.parent_component.get_primary_key_value_tuple(row)
                assert isinstance(primary_key, tuple)
            except KeyError:
                primary_key = None
            except AttributeError as e:
                raise ValueError(f"The lookup parent_component needs to be a ReadOnlyTable or related child (needs get_primary_key_value_tuple method) {e}")

            if matching_rows is None:
                matching_rows = Lookup.VERSION_COLLECTION_TYPE()
                self._cache[lk_tuple] = matching_rows

            try:
                if primary_key is None:
                    self._len += 1
                    matching_rows[self._len] = row
                elif primary_key in matching_rows:
                    if allow_update:
                        matching_rows[primary_key] = row
                    else:
                        self.log.error('Key already in lookup!')
                        self.log.error(f'Existing row = {repr(matching_rows[primary_key])}')
                        self.log.error(f'New duplicate row = {repr(row)}')
                        raise ValueError(f'Key {lk_tuple} + primary_key {primary_key} already in cache and allow_update was False')
                else:
                    self._len += 1
                    matching_rows[primary_key] = row
            except TypeError:
                raise

            # Capture memory usage snapshots
            if self._row_size is None:
                self._get_first_row_size(row)
            else:
                self.check_estimate_row_size()

    def uncache_row(self, row: Lookup.ROW_TYPES):
        if isinstance(row, tuple) or isinstance(row, list):
            raise ValueError("{}.uncache_row requires a Row not a tuple since it needs the date".format(self.__class__.__name__))
        else:
            lk_tuple = self.get_hashable_combined_key(row)
        if self._cache is not None:
            existing_rows = self._cache.get(lk_tuple, None)
            # noinspection PyUnresolvedReferences
            primary_key = self.parent_component.get_primary_key_value_tuple(row)
            assert isinstance(primary_key, tuple)

            if existing_rows:
                # Look for and remove existing instance that are exactly the same date
                try:
                    del existing_rows[primary_key]
                except (KeyError, ValueError):
                    # Not found, that's OK
                    pass

    def uncache_set(self, row: Lookup.ROW_TYPES):
        if self._cache is not None:
            if isinstance(row, self._hashable_key_type):
                lk_tuple = row
            else:
                lk_tuple = self.get_hashable_combined_key(row)
            del self._cache[lk_tuple]

    def __iter__(self):
        """
        The natural keys will come out in any order. However, the versions within a natural key set will come out in ascending order.
        """
        if self._cache is not None:
            for row_list in list(self._cache.values()):
                for row in row_list.values():
                    yield row

    def find_in_cache(self, row: Lookup.ROW_TYPES, **kwargs) -> Row:
        """
        Find an existing row in the cache effective on the date provided.
        Can raise ValueError if the cache is not setup.
        Can raise NoResultFound if the key is not in the cache.
        Can raise BeforeAllExisting is the effective date provided is before all existing records.
        """
        matches = self.find_matches_in_cache(row, **kwargs)
        if len(matches) != 1:
            raise ValueError(f"Multiple matches ({len(matches)}) found for find_in_cache. Matches are {matches}")
        else:
            return matches[0]

    def find_matches_in_cache(self, row: Lookup.ROW_TYPES, **kwargs) -> typing.Sequence[Row]:
        """
        Find an existing row in the cache effective on the date provided.
        Can raise ValueError if the cache is not setup.
        Can raise NoResultFound if the key is not in the cache.
        Can raise BeforeAllExisting is the effective date provided is before all existing records.
        """
        assert self.cache_enabled, f"Cache not enabled for {self}"
        assert self._cache is not None, f"Cache not created for {self}"
        if isinstance(row, self._hashable_key_type):
            lk_tuple = row
            primary_key = None
        elif isinstance(row, list) or isinstance(row, tuple):
            lk_tuple = self._hashable_key_type(row)
            primary_key = None
        else:
            lk_tuple = self.get_hashable_combined_key(row)
            try:
                # noinspection PyUnresolvedReferences
                primary_key = self.parent_component.get_primary_key_value_tuple(row)
            except AttributeError as e:
                raise ValueError(f"The lookup parent_component needs to be a ReadOnlyTable or related child (needs get_primary_key_value_tuple method) {e}")
        matching_rows = self._cache.get(lk_tuple, None)
        if matching_rows is None:
            raise NoResultFound()

        if primary_key is None:
            return list(matching_rows.values())
        else:
            try:
                return [matching_rows[primary_key]]
            except KeyError as e:
                msg = str(e)
                msg += "\nmatching_rows = {}".format(matching_rows)
                # noinspection PyTypeChecker
                msg += "\nmatching_rows key[0]= {}".format(next(matching_rows.keys()))
                msg += "\nprimary_key = {}".format(primary_key)
                raise TypeError(msg)

    def find_in_remote_table(self, row: Lookup.ROW_TYPES, **kwargs) -> Row:
        """
        Find a matching row in the lookup based on the lookup index (keys)

        Only works if parent_component is based on bi_etl.components.readonlytable
        """
        raise NotImplementedError()

    def find_versions_list_in_remote_table(self, row: Lookup.ROW_TYPES) -> typing.Sequence[Row]:
        """
        Find a matching row in the lookup based on the lookup index (keys)

        Only works if parent_component is based on bi_etl.components.readonlytable
        """
        self.stats.timer.start()
        if self._remote_lookup_stmt is None:
            # noinspection PyUnresolvedReferences
            stmt = self.parent_component.select()
            stmt = super()._add_remote_stmt_where_clause(stmt)
            stmt = stmt.order_by(self.parent_component.primary_key)
            self._remote_lookup_stmt = stmt.compile()

        values_dict = super()._get_remote_stmt_where_values(row)

        # noinspection PyUnresolvedReferences
        result = list(self.parent_component.execute(self._remote_lookup_stmt, values_dict))
        rows = list()
        for result_proxy_row in result:
            row = self.parent_component.Row(data=result_proxy_row)
            rows.append(row)
        self.stats.timer.stop()
        if len(rows) == 0:
            raise NoResultFound()
        else:
            return rows



import logging
import os
import subprocess
import tempfile
import textwrap
from datetime import date, datetime
from pprint import pformat

from config_wrangler.config_templates.config_hierarchy import ConfigHierarchy
from config_wrangler.config_types.path_types import ExecutablePath
from sqlalchemy.dialects.mssql import BIT

from bi_etl.components.table import Table

log = logging.getLogger('etl.utils.bcp_helpers')


class BCPError(Exception):
    pass


class BCP_Config(ConfigHierarchy):
    path_to_bcp_exectable: ExecutablePath = 'bcp'
    batch_size: int = 10000


def create_bcp_format_file(table: Table, bcp_format_path, encoding=None, delimiter=None, row_terminator=None):
    with open(bcp_format_path, "w", encoding="utf-8") as bcp_fmt:
        field_list = list()
        column_list = list()
        max_col = len(table.columns)
        if delimiter is None:
            if encoding == 'utf_16_le':
                delimiter = '|\\0'
            else:
                delimiter = '|'
        for col_num, column in enumerate(table.columns):
            size = None
            scale = None

            c_type = column.type
            try:
                p_type = c_type.python_type
            except NotImplementedError:
                if isinstance(c_type, BIT):
                    p_type = bool
                else:
                    raise ValueError("Unexpected data type {} for column".format(c_type, column))

            # Types
            # https://docs.microsoft.com/en-us/sql/relational-databases/import-export/xml-format-files-sql-server?view=sql-server-2017
            # https://msdn.microsoft.com/en-us/library/ff718877(v=sql.105).aspx

            if p_type == int:
                size = 24
            elif p_type == bool:
                size = 1
            elif p_type == datetime:
                size = 48
            elif p_type == date:
                size = 22
            elif p_type == str:
                if c_type.length is not None:
                    if encoding == 'utf_16_le':
                        size = c_type.length * 2
                    else:
                        size = c_type.length
                    if size > 2 ** 15:
                        size = None

            if col_num + 1 == max_col:
                if row_terminator is None:
                    if encoding == 'utf_16_le':
                        delimiter = "\\r\\0\\n\\0"
                    else:
                        delimiter = "\\r\\n"
                else:
                    delimiter = row_terminator
            if encoding == 'utf_16_le':
                field_type = 'NCharTerm'
            else:
                field_type = 'CharTerm'
            field_spec = '<FIELD ID="{col_num}" xsi:type="{type}" COLLATION="" TERMINATOR="{term}"'.format(
                col_num=col_num + 1,
                type=field_type,
                term=delimiter,
            )
            if size is not None:
                field_spec += ' MAX_LENGTH="{size}"'.format(size=size)
            if scale is not None:
                field_spec += ' SCALE="{scale}"'.format(scale=scale)
            field_spec += "/>"
            field_list.append(field_spec)
            column_list.append(
                '<COLUMN SOURCE="{col_num}" NAME="{name}" />'.format(
                    col_num=col_num + 1,
                    name=column.name,
                )
            )
        bcp_fmt.write(textwrap.dedent("""\
                        <?xml version="1.0"?>
                        <BCPFORMAT xmlns="http://schemas.microsoft.com/sqlserver/2004/bulkload/format" 
                                   xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
                        <RECORD>
                        """))
        bcp_fmt.write('\n'.join(field_list))
        bcp_fmt.write(textwrap.dedent("""\n</RECORD>\n<ROW>"""))
        bcp_fmt.write('\n'.join(column_list))
        bcp_fmt.write(textwrap.dedent("""\n</ROW>\n</BCPFORMAT>"""))


def run_bcp(
        config: BCP_Config,
        table_name: str,
        file_path: str,
        database_bind,
        format_file_path=None,
        direction='in',
        delimiter='^K',
        temp_dir=None,
        start_line=1,
        encoding=None,
        ):
    if temp_dir is None:
        cleanup_temp = True
        temp_dir_obj = tempfile.TemporaryDirectory(ignore_cleanup_errors=True)
        temp_dir = temp_dir_obj.name
    else:
        temp_dir_obj = None
        cleanup_temp = False

    bcp_errors = os.path.join(temp_dir, "bcp.errors")

    if encoding == 'utf_16_le':
        char_encoding = 'UTF-16'
    else:
        # char_encoding = 'UTF-8'
        char_encoding = '65001'

    cmd = [config.path_to_bcp_exectable,
           table_name,
           # in / out
           direction,
           file_path,
           '-S', database_bind.url.host,
           '-d', database_bind.url.database,
           '-U', database_bind.url.username,
           '-P', database_bind.url.password,
           # Max errors
           # '-m', '1000',
           # '-o', bcp_output,
           '-e', bcp_errors,
           # Batch size
           '-b', str(config.batch_size),
           # encoding eg UTF-8
           '-C', char_encoding,
           # hints
           # '-h', 'CHECK_CONSTRAINTS,TABLOCK',
           # '-h', 'CHECK_CONSTRAINTS',
           # packet size = Max
           # '-a', '65535'
           ]

    # Specify start line defaults to 1 in params
    if start_line > 1:
        cmd.extend(['-F', str(start_line)])

    if format_file_path:
        cmd.extend(['-f', format_file_path])
    else:
        # UTF Mode
        if encoding == 'utf_16_le':
            cmd.extend(['-w'])
        else:
            cmd.extend(['-c'])
        cmd.extend(['-t', delimiter])

    log.debug(" ".join(cmd).replace(database_bind.url.password, '****'))
    messages = list()

    try:
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, bufsize=1, stderr=subprocess.STDOUT)
        # this did not work
        # try:
        #     outs, _ = p.communicate(input=database_bind.url.password, timeout=1)
        #     for line in outs.spltlines:
        #         if line != b'':
        #             line = line.strip()
        #             messages.append(line)
        #             log.info(line)
        # except subprocess.TimeoutExpired:
        #     pass
        while p.poll() is None:

            line = p.stdout.readline()
            if line != b'':
                line = line.strip()
                messages.append(line)
                log.info(line)
        for line in p.stdout.readlines():
            if line != b'':
                line = line.strip()
                messages.append(line)
                log.info(line)
        p.stdout.close()
        rc = p.returncode
        if rc != 0:
            log.error('bcp returned code {}'.format(rc))
        with open(bcp_errors, 'r', encoding='utf-8', errors='skip') as bcp_error_file:
            error_messages = bcp_error_file.read()
        if len(error_messages) > 0:
            if rc == 0:
                rc = 1
                log.error("Errors with rc = 0. Setting rc = 1")
            log.error('-' * 80)
            log.error('BCP error detail:')
            log.error(error_messages)
            log.error('-' * 80)

        if rc == 0:
            log.debug('BCP output parsing:')
            rows = 0
            for line in messages:
                line = line.strip()
                if line.endswith(b' rows copied.'):
                    rows = int(line[:-13])
                    log.debug("Rows message found rows = {}".format(rows))
                    break
            return rows
        else:
            raise BCPError('bcp error')

    except IOError as e:
        raise e
    except subprocess.CalledProcessError as e:
        log.error("Error code " + str(e.returncode))
        log.error('BCP messages:')
        log.error(pformat(messages))
        log.error('BCP output:')
        log.error(e.output)
        log.error('-' * 80)
        with open(bcp_errors, 'r') as bcp_error_file:
            error_messages = bcp_error_file.read()
        log.error('BCP raw errors:')
        log.error(error_messages)
        log.error('-' * 80)
        raise BCPError("BCP Error code " + str(e.returncode))
    finally:
        if cleanup_temp:
            temp_dir_obj.cleanup()


def format_value_for_bcp(value):
    if isinstance(value, datetime):
        # noinspection PyTypeChecker,PyTypeChecker
        return ('{year:4d}-{month:02d}-{day:02d} {hour:02d}:{min:02d}:{sec:06.3f}'
                .format(year=value.year,
                        month=value.month,
                        day=value.day,
                        hour=value.hour,
                        min=value.minute,
                        sec=value.second + value.microsecond / 1000000)
                )
    elif value is None:
        return ''
    else:
        return str(value).replace('|', '/')



# -*- coding: utf-8 -*-
"""
Created on Jan 5, 2016

@author: Derek Wood
"""
# https://www.python.org/dev/peps/pep-0563/
from __future__ import annotations

import gc
import typing
from datetime import datetime

import psutil

from bi_etl.components.row.row import Row
from bi_etl.config.bi_etl_config_base import BI_ETL_Config_Base
from bi_etl.exceptions import NoResultFound
from bi_etl.lookups.disk_lookup import DiskLookup
from bi_etl.lookups.lookup import Lookup
from bi_etl.timer import Timer

if typing.TYPE_CHECKING:
    from bi_etl.components.etlcomponent import ETLComponent

__all__ = ['AutoDiskLookup']


class AutoDiskLookup(Lookup):
    """
    Automatic memory / disk lookup cache.
    
    This version divides the cache into N chunks (default is 10). If RAM usage gets beyond limits, it starts moving chunks to disk.
    Once a chunk is on disk, it stays there.
    
    TODO: For use cases where the lookup will be used in a mostly sequential fashion, it would be useful to have a version that uses ranges
    instead of a hash function. Then when find_in_cache is called on a disk segment, we could swap a different segment out and bring that 
    segment in. That's a lot more complicated. We'd also want to maintain a last used date for each segment so that if we add rows to the 
    cache, we can choose the best segment to swap to disk.
    
    Also worth considering is that if we bring a segment in from disk, it would best to keep the disk version. At that point any additions 
    to that segment would need to go to both places.

    """
    def __init__(self,
                 lookup_name: str,
                 lookup_keys: list,
                 parent_component: ETLComponent,
                 config: BI_ETL_Config_Base = None,
                 use_value_cache: bool = True,
                 path=None,
                 max_percent_ram_used=None,
                 max_process_ram_usage_mb=None,
                 init_parent: bool = True,
                 **kwargs
                 ):
        if init_parent:
            super().__init__(
                lookup_name=lookup_name,
                lookup_keys=lookup_keys,
                parent_component=parent_component,
                use_value_cache=use_value_cache,
                config=config,
                )
        self._cache = None
        self.rows_cached = 0
        # First try and use passed value
        self.max_percent_ram_used = max_percent_ram_used
        # If not passed in config
        if config is not None:
            self.config = config
        else:
            self.config = parent_component.task.config
        if self.max_percent_ram_used is None:
            if self.config is not None:
                self.max_percent_ram_used = self.config.bi_etl.lookup_disk_swap_at_percent_ram_used
        # Finally default value
        if self.max_percent_ram_used is None:
            # Needs to be less than the default in bi_etl.components.table.Table.fill_cache
            self.max_percent_ram_used = 70
                        
        self.max_process_ram_usage_mb = max_process_ram_usage_mb
        # If not passed in config
        if self.max_process_ram_usage_mb is None:
            if self.config is not None:
                self.max_process_ram_usage_mb = self.config.bi_etl.lookup_disk_swap_at_process_ram_usage_mb
        # Finally default value
        if self.max_process_ram_usage_mb is None:
            self.max_process_ram_usage_mb = 2.5 * 1024**3
            
        self._set_path(path)
        
        self.ram_check_row_interval = 5000
        self.last_ram_check_at_row = 0
        self.disk_cache = None
        self.MemoryLookupClass = Lookup
        self.DiskLookupClass = DiskLookup
        if kwargs is None:
            kwargs = dict()
        kwargs['use_value_cache'] = use_value_cache
        self.lookup_class_args = kwargs
         
    def _set_path(self, path):
        if path is not None:
            self.path = path
        else:
            if self.config is not None:
                self.path = self.config.get('Cache', 'path', fallback=DiskLookup.DEFAULT_PATH)
            else:
                self.path = DiskLookup.DEFAULT_PATH

    def _init_mem_cache(self):
        self._cache = self.MemoryLookupClass(lookup_name=self.lookup_name,
                                             lookup_keys=self.lookup_keys,
                                             parent_component=self.parent_component,
                                             config=self.config,
                                             **self.lookup_class_args
                                             )
        self._cache.init_cache()
    
    def init_cache(self):
        if self.cache_enabled is None:
            self.cache_enabled = True
        if self.cache_enabled:
            self._init_mem_cache()        
    
    def get_memory_size(self):
        ram_size = 0
        if self._cache is not None:
            ram_size += self._cache.get_memory_size()
        if self.disk_cache is not None:
            ram_size += self.disk_cache.get_memory_size()
        return ram_size 
        
    def get_disk_size(self):
        disk_size = 0
        if self._cache is not None:
            disk_size += self._cache.get_disk_size()
        if self.disk_cache is not None:
            disk_size += self.disk_cache.get_disk_size()
        return disk_size     
        
    def clear_cache(self):
        if self._cache is not None:
            self._cache.clear_cache()
            del self._cache
        if self.disk_cache is not None:
            self.disk_cache.clear_cache()
            del self.disk_cache 
                    
        self._cache = None
        self.disk_cache = None
    
    def __len__(self):
        total_len = 0
        if self._cache is not None:
            total_len += len(self._cache)
        if self.disk_cache is not None:
            total_len += len(self.disk_cache) 
        return total_len

    def init_disk_cache(self):
        if self.disk_cache is None:
            self.disk_cache = self.DiskLookupClass(lookup_name=self.lookup_name,
                                                   lookup_keys=self.lookup_keys,
                                                   parent_component=self.parent_component,
                                                   config=self.config,
                                                   path=self.path,
                                                   **self.lookup_class_args
                                                   )
            self.disk_cache.init_cache()
            # Do not warn about protected access to _get_estimate_row_size
            # pylint: disable=protected-access
            self._cache.check_estimate_row_size(force_now=True)
            self.disk_cache._row_size = self._cache.estimated_row_size()
            self.disk_cache._done_get_estimate_row_size = self._cache.has_done_get_estimate_row_size()

    def flush_to_disk(self):
        if self._cache is not None and len(self._cache) > 0:
            rows_before = len(self)
            self.init_disk_cache()
            timer = Timer()
            self.log.info('Flushing {rows:,} rows to disk.'.format(rows=len(self._cache)))
            gc.collect()
            before_move_mb = self.our_process.memory_info().rss/(1024**2)
            for row in self._cache:
                self.disk_cache.cache_row(row)
            self._cache.clear_cache()
            del self._cache
            self._cache = None
            self._init_mem_cache()
            if len(self) != rows_before:
                msg = "Row count changed during flush to disk." \
                      " Rows before flush = {}, rows after flush = {}".format(rows_before, len(self))
                raise AssertionError(msg)
            self.log.info('Flushing rows took {} seconds'.format(timer.seconds_elapsed_formatted))
            gc.collect()
            after_move_mb = self.our_process.memory_info().rss/(1024**2)
            self.log.info(
                f'Flushing rows freed {before_move_mb - after_move_mb:,.3f} MB from process '
                f'before {before_move_mb:,.3f} after {after_move_mb:,.3f})'
            )

    def memory_limit_reached(self) -> bool:
        if self.max_percent_ram_used is not None:
            if psutil.virtual_memory().percent > self.max_percent_ram_used:
                self.log.warning(
                    f"{self.lookup_name} system memory limit reached "
                    f"{psutil.virtual_memory().percent} > {self.max_percent_ram_used}% "
                    f"with {self.rows_cached:,} rows of data"
                )
                return True
        process_mb = self.our_process.memory_info().rss / (1024 ** 2)
        if self.max_process_ram_usage_mb is not None:
            if process_mb > self.max_process_ram_usage_mb:
                self.log.warning("{name} process memory limit reached"
                                 " {usg:,} > {usg_limit:,} KB with {rows:,} rows of data"
                                 .format(name=self.lookup_name,
                                         rows=self.rows_cached,
                                         usg=process_mb,
                                         usg_limit=self.max_process_ram_usage_mb,
                                         )
                                 )
                return True
        return False

    def cache_row(
            self,
            row: Row,
            allow_update: bool = True,
            allow_insert: bool = True,
    ):
        """
        Adds the given row to the cache for this lookup.

        Parameters
        ----------
        row: Row
            The row to cache

        allow_update: boolean
            Allow this method to update an existing row in the cache.

        allow_insert: boolean
            Allow this method to insert a new row into the cache

        Raises
        ------
        ValueError
            If allow_update is False and an already existing row (lookup key) is passed in.

        """
        if self.cache_enabled:        
            if self._cache is None:
                self.init_cache()

            # Note: This will double count updated rows.
            # We fix it below at each ram_check_row_interval, but that calc is slower.
            self.rows_cached += 1
            
            # Note: The memory check needs to be here and not in Table.fill_cache since rows can be added to cache 
            #       during the load and not just by fill_cache.

            # Python memory is hard to free... so read first rows into RAM and then use disk for all rows after

            # Every X rows check memory limits
            if (self.disk_cache is None
                and (self.rows_cached - self.last_ram_check_at_row) >= self.ram_check_row_interval
               ):
                # Double check our cache size. Calls to cache_row might have overwritten existing rows
                self.rows_cached = len(self)
                self.last_ram_check_at_row = self.rows_cached
                if self.memory_limit_reached():
                    self.init_disk_cache()

            # Now cache the row
            if self.disk_cache is None:
                self._cache.cache_row(row, allow_update=allow_update)
            else:
                # We need to make sure each row is in only once place
                lk_tuple = self.get_hashable_combined_key(row)
                if lk_tuple in self._cache:
                    # Move existing key date ranges to disk
                    versions_collection = self.get_versions_collection(row)
                    self._cache.uncache_set(lk_tuple)
                    # Change collection type if needed
                    if not isinstance(versions_collection, DiskLookup.VERSION_COLLECTION_TYPE):
                        versions_collection = DiskLookup.VERSION_COLLECTION_TYPE(versions_collection)
                    disk_lk_tuple = self.disk_cache.get_hashable_combined_key(row)
                    self.disk_cache.cache_set(disk_lk_tuple, versions_collection)
                    # Ensure our new row is there
                    self.disk_cache.cache_row(row, allow_update=True)
                else:
                    # Cache the row to disk
                    self.disk_cache.cache_row(row, allow_update=allow_update)

    def uncache_row(self, row: Lookup.ROW_TYPES):
        if self._cache is not None:
            self._cache.uncache_row(row)
        if self.disk_cache is not None:
            self.disk_cache.uncache_row(row)

    def __iter__(self):
        """
        The rows will come out in any order.  DO NOT MODIFY cache during the loop
        """
        if self._cache is not None:
            for row in self._cache:
                yield row
        if self.disk_cache is not None:
            for row in self.disk_cache:
                yield row

    def get_versions_collection(
            self,
            row: Lookup.ROW_TYPES
            ) -> typing.MutableMapping[datetime, Row]:
        """
        This method exists for compatibility with range caches

        Parameters
        ----------
        row
            The row with keys to search row

        Returns
        -------
        A MutableMapping of rows
        """
        if not self.cache_enabled:
            raise ValueError("Lookup {} cache not enabled".format(self.lookup_name))
        if self._cache is None:
            self.init_cache()

        try:
            return self._cache.get_versions_collection(row)
        except NoResultFound:
            if self.disk_cache is not None:
                return self.disk_cache.get_versions_collection(row)
            else:
                raise NoResultFound()

    def report_on_value_cache_effectiveness(self, lookup_name: str = None):
        if lookup_name is None:
            lookup_name = self.lookup_name
        if self._cache:
            self._cache.report_on_value_cache_effectiveness(f'{lookup_name} RAM')
        if self.disk_cache:
            self.disk_cache.report_on_value_cache_effectiveness(f'{lookup_name} DISK')



"""
Created on Mar 13, 2014

@author: Derek Wood
"""

import sys


def multi_choice_question(question, choices, default=None):
    while True:
        sys.stdout.write(question)
        choice = input().lower()
        choices_lower = [c.lower() for c in choices]
        if default is not None and default not in choices:
            raise ValueError("invalid default answer: '%s'" % default)
        if default is not None and choice == '':
            return default
        elif choice in choices_lower:
            return choice
        else:
            best_candidate = None
            possible_matches = 0
            for candidate in choices:
                if choice == candidate[:len(choice)].lower():
                    best_candidate = candidate
                    possible_matches += 1
            if possible_matches == 1:
                return best_candidate
            else:
                print(("Please respond with one of {}.  Got {}".format(choices, choice)))


def yes_no(
    question: str,
    default: str = "yes"
):
    """
    Ask a yes/no question via raw_input() and return their answer.

    Parameters
    ----------
    question:
        String that is presented to the user.
    default:
        The presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is one of "yes" or "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' (or 'y' or 'n').\n")



# -*- coding: utf-8 -*-
"""
Created on Sep 17, 2014

@author: Derek Wood
"""
# https://www.python.org/dev/peps/pep-0563/
from __future__ import annotations

import collections.abc
import dataclasses
import warnings
from collections import namedtuple
from decimal import Decimal
from typing import *
from typing import Union, List, Iterable

from pydantic import BaseModel
from sqlalchemy.sql.schema import Column

from bi_etl.components.row.column_difference import ColumnDifference
from bi_etl.components.row.row_iteration_header import RowIterationHeader
from bi_etl.components.row.row_status import RowStatus
from bi_etl.utility import dict_to_str


class Row(MutableMapping):
    """
    Replacement for core SQL Alchemy, CSV or other dictionary based rows.
    Handles column names that are SQL Alchemy column objects.
    Keeps order of the columns (see columns_in_order)
    """

    # Using slots to fix the attributes to just these. This allows CPython to create Row objects 10% faster
    __slots__ = [
        '_data_values',
        'iteration_header',
        'status',
    ]
    NUMERIC_TYPES = {int, float, Decimal}
    RAISE_ON_NOT_EXIST_NAME = 'raise_on_not_exist'
    RowIterationHeader_Class = RowIterationHeader
    # For performance with the Column to str conversion we keep a cache of converted values

    def __init__(self,
                 iteration_header: RowIterationHeader,
                 data: Union[MutableMapping, list, namedtuple, None] = None,
                 status: Optional[RowStatus] = None,
                 allocate_space: bool = True):
        """
        Note: If data is passed here, it uses :py:meth:`bi_etl.components.row.row.Row.update` to map the data
        into the columns.  That is nicely automatic, but slower since it has to try various
        ways to read the data object.

        Fastest way would be to not pass any data values, and follow with a call to one of:

        * :py:meth:`bi_etl.components.row.row.Row.update_from_namedtuple`
        * :py:meth:`bi_etl.components.row.row.Row.update_from_dict`
        * :py:meth:`bi_etl.components.row.row.Row.update_from_row_proxy`
        * :py:meth:`bi_etl.components.row.row.Row.update_from_tuples`
        * :py:meth:`bi_etl.components.row.row.Row.update_from_dataclass`
        * :py:meth:`bi_etl.components.row.row.Row.update_from_pydantic`
        * :py:meth:`bi_etl.components.row.row.Row.update_from_values`

        """
        # Whatever we store here we need to either store on disk for a lookup,
        # or have a way of retrieving in __setstate__
        super().__init__()
        # We need to accept None for iteration_header for shelve to be efficient
        self._data_values = list()
        if isinstance(iteration_header, int):
            self.iteration_header = self.RowIterationHeader_Class.get_by_id(iteration_header)
            """
            The :py:class:`bi_etl.components.row.row_iteration_header.RowIterationHeader` instance that
            provides a shared definition of columns across many Row instances.

            NOTE: Changes to the columns, such as adding a new column, will replace the iteration_header
            of this Row.  If two or more Row's get the same change, they will all share the same new
            RowIterationHeader instance as their iteration_header value.
            """

        elif isinstance(iteration_header, tuple):
            self.iteration_header = self.RowIterationHeader_Class.get_by_process_and_id(iteration_header)
        else:
            assert isinstance(iteration_header, self.RowIterationHeader_Class), \
                f"First argument to Row needs to be RowIterationHeader type, got {type(iteration_header)}"
            self.iteration_header = iteration_header
        self.iteration_header.add_row(self)
        if allocate_space:
            self._extend_to_size(len(self.iteration_header.columns_in_order))
        self.status = status

        # Populate our data
        if data is not None:
            self.update(data)

    def __reduce__(self):
        # TODO: Experiment with different formats for performance and compactness
        # 91 bytes using pickle.HIGHEST_PROTOCOL, 86 bytes in test using default protocol
        status_value = None
        if self.status is not None:
            if isinstance(self.status, RowStatus):
                status_value = self.status.value
            else:
                status_value = self.status
        return (
            # A callable object that will be called to create the initial version of the object.
            self.__class__,

            # A tuple of arguments for the callable object.
            # An empty tuple must be given if the callable does not accept any argument
            (
                self.iteration_header.get_cross_process_iteration_header(),
                self._data_values,
                status_value
            ),
            # Optionally, the object’s state, which will be passed to the object’s __setstate__() method
            # as previously described.
            # If the object has no such method then, the value must be a dictionary
            # and it will be added to the object’s __dict__ attribute.
            None,

            # Optionally, an iterator (and not a sequence) yielding successive items.
            # These items will be appended to the object either using obj.append(item) or,
            # in batch, using obj.extend(list_of_items).

            # Optionally, an iterator (not a sequence) yielding successive key-value pairs.
            # These items will be stored to the object using obj[key] = value

            # PROTOCOL 5+ only
            # Optionally, a callable with a (obj, state) signature.
            # This callable allows the user to programmatically control
            # the state-updating behavior of a specific object,
            # instead of using obj’s static __setstate__() method.
            # If not None, this callable will have priority over obj’s __setstate__().
        )

    def __reduce_v1__(self):
        # TODO: Experiment with different formats for performance and compactness
        # 114 bytes in test
        status_value = None
        if self.status is not None:
            status_value = self.status.value
        outgoing_dict = {
            's': status_value,
            'v': self._data_values,
        }
        return (self.__class__,
                # A tuple of arguments for the callable object.
                (self.iteration_header.iteration_id,),
                # State to be passed to setstate
                outgoing_dict,
                )

    def __setstate_v1__(self, incoming_dict):
        if incoming_dict['s'] is not None:
            self.status = RowStatus(incoming_dict['s'])
        else:
            self.status = None
        # Restore column values
        self._data_values = incoming_dict['v']

    def update_from_namedtuple(self, source_data: namedtuple):
        """
        Update the row values from a ``namedtuple`` instance.
        Adds columns for any new names found.
        """
        # noinspection PyProtectedMember
        for column_specifier, value in zip(source_data._fields, source_data):
            column_name = self.iteration_header.get_column_name(column_specifier)
            self._raw_setitem(column_name, value)

    def update_from_dict(self, source_dict: dict):
        """
        Update the row values from a ``dict`` instance.
        Adds columns for any new names found.
        """
        for column_specifier, value in source_dict.items():
            column_name = self.iteration_header.get_column_name(column_specifier)
            self._raw_setitem(column_name, value)

    def update_from_row_proxy(self, source_row: Row):
        """
        Update the row values from a SQL Alchemy result row instance.
        Adds columns for any new names found.
        """
        for column_specifier, value in source_row.items():
            column_name = self.iteration_header.get_column_name(column_specifier)
            self._raw_setitem(column_name, value)

    def update_from_tuples(self, tuples_list: List[tuple]):
        """
        Update the row values from a ``list`` of ``tuples``.

        Each tuple should have 2 values:
          1. Column name
          2. Column value

        Adds columns for any new names found.
        """
        for column_specifier, value in tuples_list:
            column_name = self.iteration_header.get_column_name(column_specifier)
            self._raw_setitem(column_name, value)

    def update_from_dataclass(self, dataclass_inst):
        """
        Update the row values from a ``dataclass`` instance.
        Adds columns for any new names found.
        """
        self.update_from_dict(dataclass_inst.__dict__)

    def update_from_pydantic(self, pydantic_inst: BaseModel):
        """
        Update the row values from a ``pydantic`` instance of ``BaseModel``.
        Adds columns for any new names found.
        """
        # Internally pydantic __iter__ uses __dict__ but is a bit more complex
        # So going straight to __dict__ is faster
        self.update_from_dict(pydantic_inst.__dict__)

    def update_from_values(self, values_list: list):
        """
        Update the row from a list of values.
        The length of the list should be at least as long as the
        number of columns (un-filled columns will be null).
        Extra values past the number of columns will be discarded.
        """
        header_col_cnt = len(self.columns_in_order)
        self._data_values = values_list[:header_col_cnt]
        dv_col_cnt = len(self._data_values)
        if dv_col_cnt < header_col_cnt:
            self._data_values.extend([None] * (header_col_cnt - dv_col_cnt))

    def update(self, *args, **key_word_arguments):
        """
        Update the row values from a ``dict`` instance.
        Adds columns for any new names found.

        NOTE: This method is easy (nicely automatic) to use but slow
        since it has to try various ways to read the data container object.

        Consider using the appropriate one of the more specific update methods
        based on the source data container.

        * :py:meth:`bi_etl.components.row.row.Row.update_from_namedtuple`
        * :py:meth:`bi_etl.components.row.row.Row.update_from_dict`
        * :py:meth:`bi_etl.components.row.row.Row.update_from_row_proxy`
        * :py:meth:`bi_etl.components.row.row.Row.update_from_tuples`
        * :py:meth:`bi_etl.components.row.row.Row.update_from_dataclass`
        * :py:meth:`bi_etl.components.row.row.Row.update_from_pydantic`
        * :py:meth:`bi_etl.components.row.row.Row.update_from_values`
        """
        if len(key_word_arguments) > 0:
            self.update_from_dict(key_word_arguments)

        for source_data in args:
            try:
                if hasattr(source_data, '_fields'):
                    self.update_from_namedtuple(source_data)
                else:
                    # Is a dict
                    self.update_from_dict(source_data)
            except AttributeError as e:
                # Not a dict or sqlalchemy Row
                if dataclasses.is_dataclass(source_data):
                    self.update_from_dataclass(source_data)
                elif hasattr(source_data, '__iter__') and not isinstance(source_data, str):
                    try:
                        source_data = list(source_data)
                        if len(source_data) > 0:
                            # List of tuples (column_name, value) or list of values
                            # (only if we have column names already)
                            if isinstance(source_data[0], tuple):
                                self.update_from_tuples(source_data)
                            else:
                                self.update_from_values(source_data)
                    except TypeError as e1:
                        try:
                            # noinspection PyProtectedMember
                            # sqlalchemy.util._collections.ImmutableProperties
                            attributes = source_data._sa_instance_state.attrs
                            for a in attributes:  # instance of sqlalchemy.orm.state.AttributeState
                                self._raw_setitem(a.key, getattr(source_data, a.key))
                        except AttributeError as e2:  # Not iterable
                            raise ValueError(
                                f"Row couldn't get set with {source_data}. "
                                f"First Error {e1}. Error when assuming SQLAlchemy ORM row object {e2})"
                                )
                else:
                    args = str(source_data)
                    raise ValueError(
                        f"Row instance couldn't be built with source type {type(args)} value={args}. Error was {e}."
                        )

    def get_column_position(self, column_specifier) -> int:
        """
        Get the ordinal column position based on a column name (str or :py:class:`sqlalchemy.sql.schema.Column`)
        """
        column_name = self.iteration_header.get_column_name(column_specifier)
        return self.iteration_header.get_column_position(column_name)

    def get_column_name(self, column_specifier, raise_on_not_exist=True):
        if column_specifier is None:
            return None
        column_name = self.iteration_header.get_column_name(column_specifier)
        if raise_on_not_exist and not self.iteration_header.has_column(column_name):
            raise KeyError("{cls} {name} has no item {column_name} it does have {cols}"
                           .format(cls=self.__class__.__name__,
                                   name=self.name,
                                   column_name=column_name,
                                   cols=self.columns_in_order
                                   )
                           )
        return column_name

    @property
    def primary_key(self):
        return self.iteration_header.primary_key

    @primary_key.setter
    def primary_key(self, value):
        self.iteration_header.primary_key = value

    def str_formatted(self):
        return dict_to_str(self)

    @property
    def name(self):
        if self.iteration_header is not None and self.iteration_header.logical_name is not None:
            return self.iteration_header.logical_name
        else:
            return None

    def __repr__(self):
        return f'{self.__class__.__name__}(name={self.name},status={self.status},primary_key={self.primary_key},\n' \
               f'{self.str_formatted()}'

    def __str__(self):
        if self.primary_key is not None:
            key_values = [(col, self.get(col, '<N/A>')) for col in self.primary_key]
            return "{name} key_values={keys} status={s}".format(name=self.name,
                                                                keys=key_values,
                                                                s=self.status
                                                                )
        else:
            cv = [self[k] for k in self.columns_in_order[:5]]
            return "{name} cols[:5]={cv} status={s}".format(name=self.name,
                                                            cv=cv,
                                                            s=self.status
                                                            )

    def __contains__(self, column_specifier):
        column_name = self.iteration_header.get_column_name(column_specifier)
        return self.iteration_header.has_column(column_name)

    def __getitem__(self, column_specifier):
        column_name = self.iteration_header.get_column_name(column_specifier)
        position = self.iteration_header.get_column_position(column_name)
        try:
            return self._data_values[position]
        except IndexError:
            return None

    def get(self, column_specifier, default_value=None):
        column_name = self.iteration_header.get_column_name(column_specifier)
        try:
            position = self.iteration_header.get_column_position(column_name)
            try:
                return self._data_values[position]
            except IndexError:
                return default_value
        except KeyError:
            return default_value

    @property
    def as_dict(self) -> dict:
        return dict(zip(self.columns_in_order, self._data_values))

    @property
    def as_key_value_list(self) -> list:
        return list(zip(self.columns_in_order, self._data_values))

    def items(self) -> collections.abc.ItemsView:
        return collections.abc.ItemsView(
            self.as_dict
        )

    def __len__(self):
        return len(self.columns_in_order)

    def __iter__(self):
        for column_name in self.columns_in_order:
            yield column_name

    def __copy__(self):
        return self.clone()

    def keys(self) -> collections.abc.KeysView:
        return collections.abc.KeysView(
            {k: None for k in self.iteration_header.columns_in_order}
        )

    def _extend_to_size(self, desired_size):
        current_length = len(self._data_values)
        if current_length < desired_size:
            self._data_values.extend([None for _ in range(desired_size - current_length)])

    def _raw_setitem(self, column_name: Union[str, Column], value):
        self.iteration_header = self.iteration_header.row_set_item(column_name, value, self)

    def __setitem__(self, key, value):
        key_name = self.iteration_header.get_column_name(key)
        self._raw_setitem(key_name, value)

    def set_keeping_parent(self, column_name: Union[str, Column], value):
        """
        Save and restore the iteration header parent in case we are adding
        the key to the header. This saves time in build_row since it can
        know the row is "safe" for quick building

        :param column_name:
        :param value:
        :return: None
        """
        current_parent = self.iteration_header.parent
        self[column_name] = value
        self.iteration_header.parent = current_parent

    def get_name_by_position(self, position):
        """
        Get the column name in a given position.
        Note: The first column position is 1 (not 0 like a python list).
        """
        assert 0 < position <= self.iteration_header.column_count, IndexError(
            "Position {} is invalid. Expected 1 to {}".format(position, self.iteration_header.column_count)
        )

        # -1 because positions are 1 based not 0 based
        # noinspection PyProtectedMember
        return self.iteration_header._columns_in_order[position - 1]

    def get_by_position(self, position):
        """
        Get the column value by position.
        Note: The first column position is 1 (not 0 like a python list).
        """
        assert 0 < position <= self.iteration_header.column_count, IndexError(
            "Position {} is invalid. Expected 1 to {}".format(position, self.iteration_header.column_count)
        )
        if position <= len(self._data_values):
            # -1 because positions are 1 based not 0 based
            return self._data_values[position - 1]
        else:
            return None

    def set_by_zposition_unsafe(self, zposition, value):
        self._data_values[zposition] = value

    def set_by_zposition(self, zposition, value):
        """
        Set the column value by zposition (zero based)
        Note: The first column position is 0 for this method
        """
        if 0 <= zposition < self.iteration_header.column_count:
            if len(self._data_values) <= zposition:
                self._extend_to_size(zposition + 1)
            self._data_values[zposition] = value
        else:
            raise IndexError(
                "zPosition {} is invalid. Expected 0 to {}".format(zposition, self.iteration_header.column_count - 1)
            )

    def set_by_position(self, position, value):
        """
        Set the column value by position.
        Note: The first column position is 1 (not 0 like a python list).
        """
        try:
            self.set_by_zposition(position-1, value)
        except IndexError:
            raise IndexError(
                "Position {} is invalid. Expected 1 to {}".format(position, self.iteration_header.column_count)
            )

    def rename_column(self, old_name, new_name, ignore_missing=False):
        """
        Rename a column

        Parameters
        ----------
        old_name: str
            The name of the column to find and rename.

        new_name: str
            The new name to give the column.

        ignore_missing: boolean
            Ignore (don't raise error) if we don't have a column with the name in old_name.
            Defaults to False
        """
        old_name = self.iteration_header.get_column_name(old_name)
        new_name = self.iteration_header.get_column_name(new_name)
        self.iteration_header = self.iteration_header.rename_column(old_name,
                                                                    new_name,
                                                                    ignore_missing=ignore_missing)

    def rename_columns(self,
                       rename_map: Union[dict, List[tuple]],
                       ignore_missing: bool = False):
        """
        Rename many columns at once.

        Parameters
        ----------
        rename_map
            A dict or list of tuples to use to rename columns.
            Note: a list of tuples is better to use if the renames need to happen in a certain order.

        ignore_missing
            Ignore (don't raise error) if we don't have a column with the name in old_name.
            Defaults to False
        """
        self.iteration_header = self.iteration_header.rename_columns(rename_map, ignore_missing=ignore_missing)

    def __delitem__(self, column_specifier):
        column_name = self.iteration_header.get_column_name(column_specifier)
        self.iteration_header = self.iteration_header.row_remove_column(column_name, self)

    def remove_columns(self,
                       remove_list,
                       ignore_missing=False):
        """
        Remove columns from this row instance (changes to a new RowIterationHeader)

        Parameters
        ----------
        remove_list:
            A list of column names to remove

        ignore_missing:
            Ignore (don't raise error) if we don't have a column with a given name
            Defaults to False
        """
        for column_specifier in remove_list:
            column_name = self.iteration_header.get_column_name(column_specifier)
            self.iteration_header = self.iteration_header.row_remove_column(column_name,
                                                                            row=self,
                                                                            ignore_missing=ignore_missing)

    def clone(self) -> 'Row':
        """
        Create a clone of this row.
        """
        # Make the new row with the same header
        sub_row = self.__class__(iteration_header=self.iteration_header)
        # Copy data
        sub_row._data_values = self._data_values.copy()
        return sub_row

    def subset(
            self,
            exclude: Optional[Iterable] = None,
            rename_map: Optional[Union[dict, List[tuple]]] = None,
            keep_only: Optional[Iterable] = None,
            ) -> 'Row':
        """
        Return a new row instance with a subset of the columns. Original row is not modified
        Excludes are done first, then renames and finally keep_only.
        New instance will have a different RowIterationHeader.

        Parameters
        ----------
        exclude:
            A list of column names (before renames) to exclude from the subset.
            Optional. Defaults to no excludes.

        rename_map:
            A dict to use to rename columns.
            Optional. Defaults to no renames.

        keep_only:
            A list of column names (after renames) of columns to keep.
            Optional. Defaults to keep all.
        """
        # Checks for clone operation
        doing_clone = True

        if keep_only is not None:
            keep_only = set([self.iteration_header.get_column_name(c) for c in keep_only])
            doing_clone = False

        if exclude is None:
            exclude = []
        else:
            exclude = set([self.iteration_header.get_column_name(c) for c in exclude])
            doing_clone = False

        if rename_map is not None:
            doing_clone = False

        if doing_clone:
            sub_row = self.clone()
        else:
            # Make a new row with new header
            sub_row = self.iteration_header.row_subset(row=self,
                                                       exclude=exclude,
                                                       rename_map=rename_map,
                                                       keep_only=keep_only)
        return sub_row

    @property
    def column_set(self) -> frozenset:
        """
        An ImmutableSet of the columns of this row.
        Used to store different row configurations in a dictionary or set.

        WARNING: The resulting set is not ordered. Do not use if the column order affects the operation.
        See positioned_column_set instead.

        Pass through call to iteration_header.column_set.
        """
        return self.iteration_header.column_set

    @property
    def column_count(self) -> int:
        """
        Returns count of how many columns are in this row.

        Pass through call to iteration_header.column_count.
        """
        return self.iteration_header.column_count

    @property
    def positioned_column_set(self) -> Set[tuple]:
        """
        An ImmutableSet of the tuples (column, position) for this row.
        Used to store different row configurations in a dictionary or set.

        Note: column_set would not always work here because the set is not ordered even though the columns are.

        Pass through call to iteration_header.positioned_column_set.
        """
        return self.iteration_header.positioned_column_set

    def column_position(self, column_name):
        """
        Get the column position (1 based) given a column name.

        Parameters
        ----------
        column_name: str
            The column name to find the position of
        """
        normalized_name = self.iteration_header.get_column_name(column_name)
        return self.columns_in_order.index(normalized_name) + 1  # index is 0 based, positions are 1 based

    @property
    def columns_in_order(self) -> Sequence:
        """
        A list of the columns of this row in the order they were defined.

        Note: If the Row was created using a dict or dict like source, there was no order for the Row to work with.
        """
        return self.iteration_header.columns_in_order

    def values(self) -> List:
        """
        Return a list of the row values in the same order as the columns.
        """
        # TODO: Change to return either ValuesView or tuple of values.
        #       tuple tests as a lot faster, but ValuesView is what callers might expect
        #       since MutableMapping returns ValuesView.
        #       In either case, the call will be slower, but safer since the caller
        #       would not be able to break the row by changing the list of values.
        #       See also values_in_order.  We also need to deprecate one of the methods.
        return self._data_values

    def _values_equal_coerce(self, val1, val2, col_name):
        if val1 is None:
            if val2 is None:
                return True
            else:
                return False
        elif val2 is None:
            return False
        elif type(val1) == type(val2):
            return val1 == val2
        elif type(val1) in self.NUMERIC_TYPES and type(val2) in self.NUMERIC_TYPES:
            return Decimal(val1) == Decimal(val2)
        else:
            msg = '{row} data type mismatch on compare of {col_name} {type1} vs {type2}'.format(
                row=self.name,
                col_name=col_name,
                type1=type(val1),
                type2=type(val2),
            )
            warnings.warn(msg)
            return str(val1) == str(val2)

    def compare_to(self,
                   other_row: 'Row',
                   exclude: Iterable = None,
                   compare_only: Iterable = None,
                   coerce_types: bool = True) -> MutableSequence[ColumnDifference]:
        """
        Compare one RowCaseInsensitive to another. Returns a list of differences.

        Parameters
        ----------
        other_row
        exclude
        compare_only
        coerce_types

        Returns
        -------
        List of differences
        """
        if compare_only is not None:
            compare_only = set([other_row.get_column_name(c, raise_on_not_exist=False) for c in compare_only])

        if exclude is None:
            exclude = []
        else:
            exclude = set([other_row.get_column_name(c, raise_on_not_exist=False) for c in exclude])

        differences_list = list()
        for other_col_name, other_col_value in other_row.items():
            if other_col_name not in exclude:
                if compare_only is None or other_col_name in compare_only:
                    existing_column_value = self[other_col_name]
                    if coerce_types:
                        values_equal = self._values_equal_coerce(existing_column_value, other_col_value, other_col_name)
                    else:
                        values_equal = (existing_column_value == other_col_value)
                    if not values_equal:
                        differences_list.append(ColumnDifference(column_name=other_col_name,
                                                                 old_value=existing_column_value,
                                                                 new_value=other_col_value,
                                                                 )
                                                )
        return differences_list

    def __eq__(self, other):
        try:
            diffs = self.compare_to(other)
        except KeyError:
            return False
        return diffs == []

    def transform(self,
                  column_specifier: str,
                  transform_function: Callable,
                  *args,
                  **kwargs):
        # noinspection PyIncorrectDocstring
        """
        Apply a transformation to a column.
        The transformation function must take the value to be transformed as it's first argument.

        Parameters
        ----------
        column_specifier: str
            The column name in the row to be transformed
        transform_function: func
            The transformation function to use. It must take the value to be transformed as it's first argument.
        args:
            Positional arguments to pass to transform_function
        kwargs:
            Keyword arguments to pass to transform_function

        Keyword Parameters Used Directly
        --------------------------------
        raise_on_not_exist:
            Should this function raise an error if the column_specifier doesn't match an existing column.
            Must be passed as a keyword arg
            Defaults to True

        All other keyword parameters are passed along to the transform_function
        """
        # noinspection PyPep8Naming

        raise_on_not_exist = True
        if Row.RAISE_ON_NOT_EXIST_NAME in kwargs:
            raise_on_not_exist = kwargs[Row.RAISE_ON_NOT_EXIST_NAME]
            del kwargs[Row.RAISE_ON_NOT_EXIST_NAME]

        try:
            column_name = self.iteration_header.get_column_name(column_specifier)
            position = self.iteration_header.get_column_position(column_name)
            value = self._data_values[position]
        except KeyError as e:
            # If we get here, everything failed
            if raise_on_not_exist:
                raise e
            return self
        try:
            new_value = transform_function(value, *args, **kwargs)
            self._data_values[position] = new_value
        except Exception as e:
            raise ValueError(f"{transform_function} on {column_name} with value {value} yielded exception {e}")

        return self

Row.values_in_order = Row.values



from pathlib import Path
from typing import Optional

from config_wrangler.config_templates.aws.s3_bucket import S3_Bucket_Folder


class S3_Bulk_Loader_Config(S3_Bucket_Folder):
    class Config:
        validate_all = True
        validate_assignment = True
        allow_mutation = True

    temp_file_path: Optional[Path] = None
    s3_files_to_generate: int = None
    s3_file_max_rows: int = None
    s3_clear_before: bool = True
    s3_clear_when_done: bool = True
    analyze_compression: str = None  # Current Redshift options PRESET, ON, OFF (or TRUE, FALSE for the latter options)

    def validate_files(self):
        if self.s3_file_max_rows is not None and self.s3_files_to_generate is not None:
            raise ValueError(f"S3_Bulk_Loader_Config can not have both s3_file_max_rows and s3_files_to_generate")

        elif self.s3_file_max_rows is None and self.s3_files_to_generate is None:
            raise ValueError(
                f"S3_Bulk_Loader_Config needs either s3_file_max_rows or s3_files_to_generate set (not both)"
            )



import inspect
from pathlib import Path


def get_package_root_path(obj: object = None) -> Path:
    """
    Get the root path of a package given an object (default is this module).

    Parameters
    ----------
    obj:
        The object to inspect for package. Defaults to this module

    Returns
    -------
        The root path of the package as a pathlib.Path
    """
    if obj is None:
        obj = get_package_root_path
    module_path = Path(inspect.getfile(obj))
    return module_path.parents[1]


def get_package_root_str(obj: object = None) -> str:
    """
    Get the root path of a package given an object (default is this module).

    Parameters
    ----------
    obj:
        The object to inspect for package. Defaults to this module

    Returns
    -------
        The root path of the package as a string
    """
    return str(get_package_root_path(obj))


if __name__ == '__main__':
    print(get_package_root_str())



"""
Created on Sep 18, 2014

@author: Derek Wood
"""
import sys
from argparse import ArgumentParser

from bi_etl.scheduler.scheduler_interface import SchedulerInterface
from bi_etl.scheduler.status import Status
from bi_etl.scheduler.task import run_task

if __name__ == '__main__':
    parser = ArgumentParser(description="Run ETL")
    parser.add_argument('--task', type=str, help='deprecated way of specifying the task' )
    parser.add_argument('--via_scheduler', action='store_true', help='Run the task via the scheduler (asynchronous unless --wait is specified)' )
    parser.add_argument('--wait', action='store_true', help='Wait for the scheduler to finish running the task (synchronous run)' )
    parser.add_argument('--config', type=str, help='path to config file or files (comma separated) (not supported for via_scheduler)' )
    parser.add_argument('--param', type=str, nargs='?', action='append', help='parameter to pass eg.: --param foo=bar \n All parameters will be passed as strings')
    parser.add_argument('tasks_to_run', type=str, nargs='*')
    args = parser.parse_args()
    
    # print(sys.argv)
    succeeded = False
    
    if args.task:
        print("task to run {} (Note: --task is deprecated)".format(args.task))
        succeeded =run_task(args.task)
    elif args.tasks_to_run:
        if args.via_scheduler:
            print("Running via scheduler")
            sched = SchedulerInterface()
        print("tasks to run {}".format(args.tasks_to_run))
        parameters = dict()
        if args.param is not None:
            print(("args.param = {}".format(args.param)))
            for parm in args.param:
                print(("parm = {}".format(parm)))
                (parm_name, parm_value) = parm.split("=")
                parameters[parm_name] = parm_value
        for task in args.tasks_to_run:
            if args.via_scheduler:
                task_id = sched.add_task_by_partial_name(partial_module_name=task, parameters=parameters)
                if args.wait:
                    status = sched.wait_for_task(task_id)
                    succeeded = status == Status.succeeded
                else:
                    succeeded = True
            else:
                succeeded = run_task(task, config=args.config, parameters=parameters)
                        
    else:
        parser.print_usage()
        succeeded = True
    
    # Exit code 0 == success so it's the inverse of succeeded
    # Make really sure to call sys.exit with an integer otherwise it doesn't work.
    if not succeeded:
        sys.exit(99)
    else:       
        sys.exit(0)



"""
Created on March 28, 2022

@author: 
"""
import logging
import warnings

import boto3
import keyring


def get_user_id(
    **kwargs
):
    raise DeprecationWarning("Replaced with self.config.my_config_section.user_id. "
                             "Most likely not needed since the config S3_Bucket will connect seamlessly")


def get_boto3_session_from_config(
        config,
        user_id: str = None,
        password: str = None,
        keyring_system: str = 's3'
):
    warnings.warn("Replace with self.s3_helper_from_config", DeprecationWarning, stacklevel=2)

    log = logging.getLogger(__name__)

    if user_id is None:
        user_id = get_user_id(config)

    if password is None:
        password = keyring.get_password(keyring_system, user_id)
        if password is None:
            raise ValueError(f"Password not supplied in keyring for {keyring_system} {user_id}")
        # Detect short passwords from bad copy/paste
        elif len(password) < 3:
            raise ValueError(f"Password too short ({password}) in keyring for {keyring_system} {user_id}")

    log.info(f'Connecting to S3 with ID {user_id}')
    session = boto3.session.Session(
        aws_access_key_id=user_id,
        aws_secret_access_key=password
    )

    return session


class Boto3_Base(object):
    SERVICE = None

    def __init__(self, session, region_name: str = None):
        self.session = session
        self.region_name = region_name
        self._client = None
        self._resource = None

    @property
    def resource(self, ):
        if self._resource is None:
            self._resource = self.session.resource(self.SERVICE, region_name=self.region_name)

        return self._resource

    @property
    def client(self, ):
        if self._client is None:
            self._client = self.session.client(self.SERVICE, region_name=self.region_name)

        return self._client
